# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

"""
This file is autogenerated by tools/library/gen_F.py
This is the __init__.py file of the library
All modules are in ./<module>.py with name class <module>
Export all <module> classes here
Do it programmatically instead of specializing each manually
This way we can add new modules without changing this file
"""

# Disable ruff warning for whole block
# flake8: noqa: F401
# flake8: noqa: I001
# flake8: noqa: E501

from faebryk.library.has_designator_prefix import has_designator_prefix
from faebryk.library.has_pcb_position import has_pcb_position
from faebryk.library.Electrical import Electrical
from faebryk.library.has_esphome_config import has_esphome_config
from faebryk.library.is_esphome_bus import is_esphome_bus
from faebryk.library.has_single_electric_reference import has_single_electric_reference
from faebryk.library.can_specialize_defined import can_specialize_defined
from faebryk.library.Power import Power
from faebryk.library.can_be_surge_protected import can_be_surge_protected
from faebryk.library.has_net_name import has_net_name
from faebryk.library.is_bus_parameter import is_bus_parameter
from faebryk.library.Signal import Signal
from faebryk.library.is_optional_defined import is_optional_defined
from faebryk.library.has_footprint import has_footprint
from faebryk.library.Mechanical import Mechanical
from faebryk.library.has_overriden_name import has_overriden_name
from faebryk.library.is_pickable import is_pickable
from faebryk.library.has_linked_pad import has_linked_pad
from faebryk.library.has_reference import has_reference
from faebryk.library.has_descriptive_properties import has_descriptive_properties
from faebryk.library.is_app_root import is_app_root
from faebryk.library.has_layout_transform import has_layout_transform
from faebryk.library.can_bridge import can_bridge
from faebryk.library.can_specialize import can_specialize
from faebryk.library.has_datasheet import has_datasheet
from faebryk.library.has_designator import has_designator
from faebryk.library.has_part_picked import has_part_picked
from faebryk.library.has_simple_value_representation import has_simple_value_representation
from faebryk.library.has_capacitance import has_capacitance
from faebryk.library.has_construction_dependency import has_construction_dependency
from faebryk.library.has_kicad_ref import has_kicad_ref
from faebryk.library.has_pcb_layout import has_pcb_layout
from faebryk.library.has_pcb_routing_strategy import has_pcb_routing_strategy
from faebryk.library.is_optional import is_optional
from faebryk.library.has_reference_layout import has_reference_layout
from faebryk.library.has_resistance import has_resistance
from faebryk.library.has_single_connection import has_single_connection
from faebryk.library.is_not_pickable import is_not_pickable
from faebryk.library.is_representable_by_single_value import is_representable_by_single_value
from faebryk.library.has_pcb_position_defined import has_pcb_position_defined
from faebryk.library.has_pcb_position_defined_relative import has_pcb_position_defined_relative
from faebryk.library.has_pcb_position_defined_relative_to_parent import has_pcb_position_defined_relative_to_parent
from faebryk.library.PJ398SM import PJ398SM
from faebryk.library.RJ45_Receptacle import RJ45_Receptacle
from faebryk.library.XtalIF import XtalIF
from faebryk.library.has_pin_association_heuristic import has_pin_association_heuristic
from faebryk.library.has_esphome_config_defined import has_esphome_config_defined
from faebryk.library.is_esphome_bus_defined import is_esphome_bus_defined
from faebryk.library.has_single_electric_reference_defined import has_single_electric_reference_defined
from faebryk.library.Filter import Filter
from faebryk.library.Logic import Logic
from faebryk.library.Footprint import Footprint
from faebryk.library.has_overriden_name_defined import has_overriden_name_defined
from faebryk.library.is_pickable_by_part_number import is_pickable_by_part_number
from faebryk.library.is_pickable_by_supplier_id import is_pickable_by_supplier_id
from faebryk.library.is_pickable_by_type import is_pickable_by_type
from faebryk.library.has_linked_pad_defined import has_linked_pad_defined
from faebryk.library.Symbol import Symbol
from faebryk.library.can_bridge_defined import can_bridge_defined
from faebryk.library.has_datasheet_defined import has_datasheet_defined
from faebryk.library.has_part_removed import has_part_removed
from faebryk.library.has_simple_value_representation_based_on_params import has_simple_value_representation_based_on_params
from faebryk.library.has_simple_value_representation_based_on_params_chain import has_simple_value_representation_based_on_params_chain
from faebryk.library.has_simple_value_representation_defined import has_simple_value_representation_defined
from faebryk.library.has_pcb_layout_defined import has_pcb_layout_defined
from faebryk.library.has_single_connection_impl import has_single_connection_impl
from faebryk.library.is_representable_by_single_value_defined import is_representable_by_single_value_defined
from faebryk.library.has_pin_association_heuristic_lookup_table import has_pin_association_heuristic_lookup_table
from faebryk.library.LogicOps import LogicOps
from faebryk.library.can_attach_to_footprint import can_attach_to_footprint
from faebryk.library.can_attach_via_pinmap import can_attach_via_pinmap
from faebryk.library.has_footprint_impl import has_footprint_impl
from faebryk.library.has_kicad_footprint import has_kicad_footprint
from faebryk.library.has_descriptive_properties_defined import has_descriptive_properties_defined
from faebryk.library.Pad import Pad
from faebryk.library.has_symbol_layout import has_symbol_layout
from faebryk.library.Button import Button
from faebryk.library.GDT import GDT
from faebryk.library.BJT import BJT
from faebryk.library.Diode import Diode
from faebryk.library.MOSFET import MOSFET
from faebryk.library.LogicGate import LogicGate
from faebryk.library.has_footprint_defined import has_footprint_defined
from faebryk.library.Net import Net
from faebryk.library.can_attach_via_pinmap_pinlist import can_attach_via_pinmap_pinlist
from faebryk.library.has_equal_pins import has_equal_pins
from faebryk.library.has_kicad_manual_footprint import has_kicad_manual_footprint
from faebryk.library.has_pcb_routing_strategy_greedy_direct_line import has_pcb_routing_strategy_greedy_direct_line
from faebryk.library.has_symbol_layout_defined import has_symbol_layout_defined
from faebryk.library.TVS import TVS
from faebryk.library.NFET import NFET
from faebryk.library.PFET import PFET
from faebryk.library.LogicGates import LogicGates
from faebryk.library.can_attach_to_footprint_symmetrically import can_attach_to_footprint_symmetrically
from faebryk.library.can_attach_to_footprint_via_pinmap import can_attach_to_footprint_via_pinmap
from faebryk.library.has_pcb_routing_strategy_manual import has_pcb_routing_strategy_manual
from faebryk.library.has_pcb_routing_strategy_via_to_layer import has_pcb_routing_strategy_via_to_layer
from faebryk.library.can_attach_via_pinmap_equal import can_attach_via_pinmap_equal
from faebryk.library.has_equal_pins_in_ifs import has_equal_pins_in_ifs
from faebryk.library.has_kicad_footprint_equal_ifs import has_kicad_footprint_equal_ifs
from faebryk.library.KicadFootprint import KicadFootprint
from faebryk.library.SurgeProtection import SurgeProtection
from faebryk.library.Crystal import Crystal
from faebryk.library.Fuse import Fuse
from faebryk.library.Inductor import Inductor
from faebryk.library.Resistor import Resistor
from faebryk.library.Switch import Switch
from faebryk.library.B4B_ZR_SM4_TF import B4B_ZR_SM4_TF
from faebryk.library.DE9Connector import DE9Connector
from faebryk.library.Header import Header
from faebryk.library.pf_533984002 import pf_533984002
from faebryk.library.DIP import DIP
from faebryk.library.QFN import QFN
from faebryk.library.SMDTwoPin import SMDTwoPin
from faebryk.library.SOIC import SOIC
from faebryk.library.has_kicad_footprint_equal_ifs_defined import has_kicad_footprint_equal_ifs_defined
from faebryk.library.Mounting_Hole import Mounting_Hole
from faebryk.library.TestPoint import TestPoint
from faebryk.library.has_explicit_part import has_explicit_part
from faebryk.library.has_package import has_package
from faebryk.library.is_surge_protected import is_surge_protected
from faebryk.library.Common_Mode_Filter import Common_Mode_Filter
from faebryk.library.Potentiometer import Potentiometer
from faebryk.library.Capacitor import Capacitor
from faebryk.library.has_package_requirement import has_package_requirement
from faebryk.library.is_surge_protected_defined import is_surge_protected_defined
from faebryk.library.CapacitorElectrolytic import CapacitorElectrolytic
from faebryk.library.Crystal_Oscillator import Crystal_Oscillator
from faebryk.library.MultiCapacitor import MultiCapacitor
from faebryk.library.is_decoupled import is_decoupled
from faebryk.library.can_be_surge_protected_defined import can_be_surge_protected_defined
from faebryk.library.can_be_decoupled import can_be_decoupled
from faebryk.library.ElectricPower import ElectricPower
from faebryk.library.B0505S_1WR3 import B0505S_1WR3
from faebryk.library.Battery import Battery
from faebryk.library.Comparator import Comparator
from faebryk.library.ElectricSignal import ElectricSignal
from faebryk.library.Fan import Fan
from faebryk.library.LED import LED
from faebryk.library.NetTie import NetTie
from faebryk.library.OpAmp import OpAmp
from faebryk.library.Regulator import Regulator
from faebryk.library.Relay import Relay
from faebryk.library.XT30PW2plus2 import XT30PW2plus2
from faebryk.library.can_be_decoupled_rails import can_be_decoupled_rails
from faebryk.library.ButtonCell import ButtonCell
from faebryk.library.DifferentialPair import DifferentialPair
from faebryk.library.ElectricLogic import ElectricLogic
from faebryk.library.FilterElectricalLC import FilterElectricalLC
from faebryk.library.FilterElectricalRC import FilterElectricalRC
from faebryk.library.PowerMux import PowerMux
from faebryk.library.ResistorVoltageDivider import ResistorVoltageDivider
from faebryk.library.Shenzhen_Kinghelm_Elec_KH_BNC75_3511 import Shenzhen_Kinghelm_Elec_KH_BNC75_3511
from faebryk.library.PANASONIC_AQY212EHAX import PANASONIC_AQY212EHAX
from faebryk.library.PoweredLED import PoweredLED
from faebryk.library.RS485HalfDuplex import RS485HalfDuplex
from faebryk.library.USB_Type_C_Receptacle_24_pin import USB_Type_C_Receptacle_24_pin
from faebryk.library.Diodes_Incorporated_AP255x_x import Diodes_Incorporated_AP255x_x
from faebryk.library.ElectricLogicGate import ElectricLogicGate
from faebryk.library.EnablePin import EnablePin
from faebryk.library.Ethernet import Ethernet
from faebryk.library.GenericBusProtection import GenericBusProtection
from faebryk.library.I2C import I2C
from faebryk.library.I2S import I2S
from faebryk.library.JTAG import JTAG
from faebryk.library.MultiSPI import MultiSPI
from faebryk.library.Pinmux import Pinmux
from faebryk.library.RS232 import RS232
from faebryk.library.SK6812 import SK6812
from faebryk.library.SK9822_EC20 import SK9822_EC20
from faebryk.library.SNx4LVC541A import SNx4LVC541A
from faebryk.library.SPI import SPI
from faebryk.library.SWD import SWD
from faebryk.library.Sercom import Sercom
from faebryk.library.TXS0102DCUR import TXS0102DCUR
from faebryk.library.UART_Base import UART_Base
from faebryk.library.USB2_0_IF import USB2_0_IF
from faebryk.library.Wuxi_I_core_Elec_AiP74LVC1T45GB236_TR import Wuxi_I_core_Elec_AiP74LVC1T45GB236_TR
from faebryk.library.XL_3528RGBW_WS2812B import XL_3528RGBW_WS2812B
from faebryk.library.can_switch_power import can_switch_power
from faebryk.library.pf_74AHCT2G125 import pf_74AHCT2G125
from faebryk.library.TPS2116 import TPS2116
from faebryk.library.Resistor_Voltage_Divider import Resistor_Voltage_Divider
from faebryk.library.ElecSuper_PSM712_ES import ElecSuper_PSM712_ES
from faebryk.library.Diodes_Incorporated_AP2552W6_7 import Diodes_Incorporated_AP2552W6_7
from faebryk.library.Diodes_Incorporated_AP2553W6_7 import Diodes_Incorporated_AP2553W6_7
from faebryk.library.ElectricLogicGates import ElectricLogicGates
from faebryk.library.Logic74xx import Logic74xx
from faebryk.library.LDO import LDO
from faebryk.library.BH1750FVI_TR import BH1750FVI_TR
from faebryk.library.EEPROM import EEPROM
from faebryk.library.HDMI import HDMI
from faebryk.library.INA228 import INA228
from faebryk.library.ISO1540 import ISO1540
from faebryk.library.M24C08_FMN6TP import M24C08_FMN6TP
from faebryk.library.OLED_Module import OLED_Module
from faebryk.library.QWIIC_Connector import QWIIC_Connector
from faebryk.library.SCD40 import SCD40
from faebryk.library.SPIFlash import SPIFlash
from faebryk.library.RP2040Pinmux import RP2040Pinmux
from faebryk.library.DE9RS232Connector import DE9RS232Connector
from faebryk.library.SWDConnector import SWDConnector
from faebryk.library.Analog_Devices_ADM2587EBRWZ import Analog_Devices_ADM2587EBRWZ
from faebryk.library.HLK_LD2410B_P import HLK_LD2410B_P
from faebryk.library.PM1006 import PM1006
from faebryk.library.TD541S485H import TD541S485H
from faebryk.library.TXS0102DCUR_UART import TXS0102DCUR_UART
from faebryk.library.UART import UART
from faebryk.library.UART_RS485 import UART_RS485
from faebryk.library.USB2514B import USB2514B
from faebryk.library.USB2_0 import USB2_0
from faebryk.library.USB3_IF import USB3_IF
from faebryk.library.USB_Type_C_Receptacle_16_pin import USB_Type_C_Receptacle_16_pin
from faebryk.library.WCHJiangsu_Qin_Heng_CH224K import WCHJiangsu_Qin_Heng_CH224K
from faebryk.library.can_switch_power_defined import can_switch_power_defined
from faebryk.library.TPS2116_ReferenceDesign import TPS2116_ReferenceDesign
from faebryk.library.RS485_Bus_Protection import RS485_Bus_Protection
from faebryk.library.CD4011 import CD4011
from faebryk.library.ME6211C33M5G_N import ME6211C33M5G_N
from faebryk.library.INA228_ReferenceDesign import INA228_ReferenceDesign
from faebryk.library.ISO1540_ReferenceDesign import ISO1540_ReferenceDesign
from faebryk.library.Winbond_Elec_W25Q128JVSIQ import Winbond_Elec_W25Q128JVSIQ
from faebryk.library.Winbond_W25Q16JVUXIQ import Winbond_W25Q16JVUXIQ
from faebryk.library.RP2040 import RP2040
from faebryk.library.Analog_Devices_ADM2587EBRWZ_ReferenceDesign import Analog_Devices_ADM2587EBRWZ_ReferenceDesign
from faebryk.library.CH344 import CH344
from faebryk.library.SP3243E import SP3243E
from faebryk.library.CBM9002A_56ILG import CBM9002A_56ILG
from faebryk.library.CH340x import CH340x
from faebryk.library.CH342 import CH342
from faebryk.library.ESP32_C3 import ESP32_C3
from faebryk.library.MCP2221A import MCP2221A
from faebryk.library.USB2_0_ESD_Protection import USB2_0_ESD_Protection
from faebryk.library.USBLC6_2P6 import USBLC6_2P6
from faebryk.library.USB_Type_C_Receptacle_14_pin_Vertical import USB_Type_C_Receptacle_14_pin_Vertical
from faebryk.library.USB3 import USB3
from faebryk.library.PowerSwitch import PowerSwitch
from faebryk.library.TI_CD4011BE import TI_CD4011BE
from faebryk.library.RP2040_ReferenceDesign import RP2040_ReferenceDesign
from faebryk.library.CH344Q import CH344Q
from faebryk.library.SP3243E_ReferenceDesign import SP3243E_ReferenceDesign
from faebryk.library.CBM9002A_56ILG_ReferenceDesign import CBM9002A_56ILG_ReferenceDesign
from faebryk.library.USB_RS485 import USB_RS485
from faebryk.library.CH342F import CH342F
from faebryk.library.CH342K import CH342K
from faebryk.library.ESP32_C3_MINI_1 import ESP32_C3_MINI_1
from faebryk.library.TECH_PUBLIC_USBLC6_2SC6 import TECH_PUBLIC_USBLC6_2SC6
from faebryk.library.USB_C_PSU_Vertical import USB_C_PSU_Vertical
from faebryk.library.USB3_connector import USB3_connector
from faebryk.library.USB_C import USB_C
from faebryk.library.PowerSwitchMOSFET import PowerSwitchMOSFET
from faebryk.library.PowerSwitchStatic import PowerSwitchStatic
from faebryk.library.CH342F_ReferenceDesign import CH342F_ReferenceDesign
from faebryk.library.CH342K_ReferenceDesign import CH342K_ReferenceDesign
from faebryk.library.ESP32_C3_MINI_1_ReferenceDesign import ESP32_C3_MINI_1_ReferenceDesign
from faebryk.library.USB_C_5V_PSU import USB_C_5V_PSU
from faebryk.library.USB_C_PowerOnly import USB_C_PowerOnly
from faebryk.library.Powered_Relay import Powered_Relay
from faebryk.library.LEDIndicator import LEDIndicator
from faebryk.library.CH344Q_ReferenceDesign import CH344Q_ReferenceDesign
from faebryk.library.RaspberryPiPicoBase_ReferenceDesign import RaspberryPiPicoBase_ReferenceDesign
from faebryk.library.USB2514B_ReferenceDesign import USB2514B_ReferenceDesign
from faebryk.library.WCHJiangsu_Qin_Heng_CH224K_ReferenceDesign import WCHJiangsu_Qin_Heng_CH224K_ReferenceDesign
from faebryk.library.RaspberryPiPico import RaspberryPiPico



import os
import shutil
import sys
from pathlib import Path
from subprocess import CalledProcessError

import git
import pathvalidate
import pytest

from faebryk.libs.util import repo_root as _repo_root
from faebryk.libs.util import robustly_rm_dir, run_live


class CloneError(Exception):
    """Failed to clone the repository."""


class InstallError(Exception):
    """Failed to install the project."""


class BuildError(Exception):
    """Failed to build the project."""


def build_project(prj_path: Path, request: pytest.FixtureRequest):
    """Generically "build" the project."""
    friendly_node_name = pathvalidate.sanitize_filename(str(request.node.name))
    artifact_dir = _repo_root() / "artifacts"
    profile_path = artifact_dir / (friendly_node_name + "-profile.html")

    try:
        run_live(
            [
                sys.executable,
                "-m",
                "pyinstrument",
                "--html",
                "-o",
                profile_path,
                "-i",
                "0.01",
                "-m",
                "atopile",
                "-v",
                "build",
                "--keep-picked-parts",
                "--keep-net-names",
            ],
            env={**os.environ, "NONINTERACTIVE": "1"},
            cwd=prj_path,
            stdout=print,
            stderr=print,
        )
    except CalledProcessError as ex:
        artifact_path = artifact_dir / friendly_node_name
        if artifact_path.exists():
            robustly_rm_dir(artifact_path)
        artifact_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copytree(prj_path, artifact_path, ignore=shutil.ignore_patterns(".git"))

        # Translate the error message to clearly distinguish from clone errors
        raise BuildError from ex


@pytest.mark.slow
@pytest.mark.regression
@pytest.mark.parametrize(
    "repo_uri",
    [
        ("https://github.com/atopile/spin-servo-drive"),
        ("https://github.com/atopile/esp32-s3"),
        ("https://github.com/atopile/cell-sim"),
        ("https://github.com/atopile/hil"),
        ("https://github.com/atopile/rp2040"),
        ("https://github.com/atopile/tca9548apwr"),
        ("https://github.com/atopile/nau7802"),
        ("https://github.com/atopile/lv2842xlvddcr"),
        ("https://github.com/atopile/bq24045dsqr"),
    ],
)
def test_projects(
    repo_uri: str,
    tmp_path: Path,
    request: pytest.FixtureRequest,
):
    # Clone the repository
    # Using gh to use user credentials if run locally
    try:
        run_live(
            ["gh", "repo", "clone", repo_uri, "project", "--", "--depth", "1"],
            cwd=tmp_path,
            stdout=print,
            stderr=print,
        )
    except CalledProcessError as ex:
        raise CloneError from ex

    prj_path = tmp_path / "project"

    # Generically "install" the project
    try:
        run_live(
            [sys.executable, "-m", "atopile", "-v", "install"],
            env={**os.environ, "NONINTERACTIVE": "1"},
            cwd=prj_path,
            stdout=print,
            stderr=print,
        )
    except CalledProcessError as ex:
        # Translate the error message to clearly distinguish from clone errors
        raise InstallError from ex

    build_project(prj_path, request=request)

    repo = git.Repo(prj_path)
    if any(item.a_path.endswith(".kicad_pcb") for item in repo.index.diff(None)):
        pytest.xfail("can't build with --frozen yet")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F

logger = logging.getLogger(__name__)


class is_surge_protected_defined(F.is_surge_protected.impl()):
    def __init__(self, protection: F.SurgeProtection) -> None:
        super().__init__()
        self.protection = protection

    def get_protection(self):
        return self.protection



from pathlib import Path
from textwrap import dedent

import pytest

import faebryk.core.parameter as fab_param
import faebryk.library._F as F
from atopile import errors
from atopile.datatypes import Ref
from atopile.front_end import Bob, _has_ato_cmp_attrs
from atopile.parse import parse_text_as_file
from faebryk.libs.library import L
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.util import cast_assert


def test_empty_module_build(bob: Bob):
    text = dedent(
        """
        module A:
            pass
        """
    )
    tree = parse_text_as_file(text)
    node = bob.build_ast(tree, Ref(["A"]))
    assert isinstance(node, L.Module)
    assert isinstance(node, bob.modules[":A"])


def test_simple_module_build(bob: Bob):
    text = dedent(
        """
        module A:
            a = 1
        """
    )
    tree = parse_text_as_file(text)
    node = bob.build_ast(tree, Ref(["A"]))
    assert isinstance(node, L.Module)

    param = node.runtime["a"]
    assert isinstance(param, fab_param.ParameterOperatable)
    # TODO: check value


def test_arithmetic(bob: Bob):
    text = dedent(
        """
        module A:
            a = 1 to 2 * 3
            b = a + 4
        """
    )
    tree = parse_text_as_file(text)
    node = bob.build_ast(tree, Ref(["A"]))
    assert isinstance(node, L.Module)

    # TODO: check output
    # Requires params solver to be sane


def test_simple_new(bob: Bob):
    text = dedent(
        """
        component SomeComponent:
            signal a

        module A:
            child = new SomeComponent
        """
    )

    tree = parse_text_as_file(text)
    node = bob.build_ast(tree, Ref(["A"]))

    assert isinstance(node, L.Module)
    child = Bob.get_node_attr(node, "child")
    assert child.has_trait(_has_ato_cmp_attrs)

    a = Bob.get_node_attr(child, "a")
    assert isinstance(a, F.Electrical)


def test_nested_nodes(bob: Bob):
    text = dedent(
        """
        interface SomeInterface:
            signal d
            signal e

        component SomeComponent:
            pin A1
            signal a
            a ~ A1
            signal b ~ pin 2
            signal c ~ pin "C3"

        module SomeModule:
            cmp = new SomeComponent
            intf = new SomeInterface

        module ChildModule from SomeModule:
            signal child_signal

        module A:
            child = new ChildModule
            intf = new SomeInterface
            intf ~ child.intf
        """
    )

    tree = parse_text_as_file(text)
    node = bob.build_ast(tree, Ref(["A"]))

    assert isinstance(node, L.Module)


def test_resistor(bob: Bob, repo_root: Path):
    bob.search_paths.append(
        repo_root / "test" / "common" / "resources" / ".ato" / "modules"
    )

    text = dedent(
        """
        from "generics/resistors.ato" import Resistor

        component ResistorB from Resistor:
            footprint = "R0805"

        module A:
            r1 = new ResistorB
        """
    )

    tree = parse_text_as_file(text)
    node = bob.build_ast(tree, Ref(["A"]))

    assert isinstance(node, L.Module)

    r1 = Bob.get_node_attr(node, "r1")
    assert r1.get_trait(F.has_package)._enum_set == {F.has_package.Package.R0805}


def test_standard_library_import(bob: Bob):
    text = dedent(
        """
        import Resistor
        from "interfaces.ato" import PowerAC

        module A:
            r1 = new Resistor
            power_in = new PowerAC
        """
    )

    tree = parse_text_as_file(text)
    node = bob.build_ast(tree, Ref(["A"]))

    assert isinstance(node, L.Module)

    r1 = Bob.get_node_attr(node, "r1")
    assert isinstance(r1, F.Resistor)

    assert Bob.get_node_attr(node, "power_in")


@pytest.mark.parametrize(
    "import_stmt,class_name,pkg_str,pkg",
    [
        ("import Resistor", "Resistor", "R0402", F.has_package.Package.R0402),
        (
            "from 'generics/resistors.ato' import Resistor",
            "Resistor",
            "0402",
            F.has_package.Package.R0402,
        ),
        (
            "from 'generics/capacitors.ato' import Capacitor",
            "Capacitor",
            "0402",
            F.has_package.Package.C0402,
        ),
    ],
)
def test_reserved_attrs(
    bob: Bob,
    import_stmt: str,
    class_name: str,
    pkg_str: str,
    pkg: F.has_package.Package,
    repo_root: Path,
):
    bob.search_paths.append(
        repo_root / "test" / "common" / "resources" / ".ato" / "modules"
    )

    text = dedent(
        f"""
        {import_stmt}

        module A:
            a = new {class_name}
            a.package = "{pkg_str}"
            a.mpn = "1234567890"
        """
    )

    tree = parse_text_as_file(text)
    node = bob.build_ast(tree, Ref(["A"]))

    assert isinstance(node, L.Module)

    a = Bob.get_node_attr(node, "a")
    assert a.get_trait(F.has_package)._enum_set == {pkg}
    assert a.get_trait(F.has_descriptive_properties).get_properties() == {
        DescriptiveProperties.partno: "1234567890"
    }


def test_import_ato(bob: Bob, tmp_path):
    tmp_path = Path(tmp_path)
    some_module_search_path = tmp_path / "path"
    some_module_path = some_module_search_path / "to" / "some_module.ato"
    some_module_path.parent.mkdir(parents=True)

    some_module_path.write_text(
        dedent(
            """
        import Resistor

        module SpecialResistor from Resistor:
            footprint = "R0805"
        """
        )
    )

    top_module_content = dedent(
        """
        from "to/some_module.ato" import SpecialResistor

        module A:
            r1 = new SpecialResistor
        """
    )

    bob.search_paths.append(some_module_search_path)

    tree = parse_text_as_file(top_module_content)
    node = bob.build_ast(tree, Ref(["A"]))

    assert isinstance(node, L.Module)

    r1 = Bob.get_node_attr(node, "r1")
    assert isinstance(r1, F.Resistor)


@pytest.mark.parametrize(
    "module,count", [("A", 1), ("B", 3), ("C", 5), ("D", 6), ("E", 6)]
)
def test_traceback(bob: Bob, module: str, count: int):
    text = dedent(
        """
        module A:
            doesnt_exit ~ notta_connectable

        module B:
            a = new A

        module C:
            b = new B

        module D from C:
            pass

        module E from D:
            pass
        """
    )

    tree = parse_text_as_file(text)

    with pytest.raises(errors.UserKeyError) as e:
        bob.build_ast(tree, Ref([module]))

    assert e.value.traceback is not None
    assert len(e.value.traceback) == count


# TODO: test connect
# - signal ~ signal
# - higher-level mif
# - duck-typed
def _get_mif(node: L.Node, name: str) -> L.ModuleInterface:
    return cast_assert(L.ModuleInterface, Bob.get_node_attr(node, name))


def test_signal_connect(bob: Bob):
    text = dedent(
        """
        module App:
            signal a
            signal b
            signal c
            a ~ b
        """
    )

    tree = parse_text_as_file(text)
    node = bob.build_ast(tree, Ref(["App"]))

    assert isinstance(node, L.Module)

    a = _get_mif(node, "a")
    b = _get_mif(node, "b")
    c = _get_mif(node, "c")

    assert a.is_connected_to(b)
    assert not a.is_connected_to(c)


def test_interface_connect(bob: Bob):
    text = dedent(
        """
        interface SomeInterface:
            signal one
            signal two

        module App:
            a = new SomeInterface
            b = new SomeInterface
            c = new SomeInterface
            a ~ b
        """
    )

    tree = parse_text_as_file(text)
    node = bob.build_ast(tree, Ref(["App"]))

    assert isinstance(node, L.Module)

    a = _get_mif(node, "a")
    b = _get_mif(node, "b")
    c = _get_mif(node, "c")

    assert a.is_connected_to(b)
    assert not a.is_connected_to(c)

    a_one = _get_mif(a, "one")
    b_one = _get_mif(b, "one")
    c_one = _get_mif(c, "one")
    a_two = _get_mif(a, "two")
    b_two = _get_mif(b, "two")
    c_two = _get_mif(c, "two")

    assert a_one.is_connected_to(b_one)
    assert a_two.is_connected_to(b_two)
    assert not any(
        a_one.is_connected_to(other) for other in [a_two, b_two, c_one, c_two]
    )
    assert not any(
        a_two.is_connected_to(other) for other in [a_one, b_one, c_one, c_two]
    )


def test_duck_type_connect(bob: Bob):
    text = dedent(
        """
        interface SomeInterface:
            signal one
            signal two

        interface SomeOtherInterface:
            signal one
            signal two

        module App:
            a = new SomeInterface
            b = new SomeOtherInterface
            a ~ b
        """
    )

    tree = parse_text_as_file(text)
    node = bob.build_ast(tree, Ref(["App"]))

    assert isinstance(node, L.Module)

    a = _get_mif(node, "a")
    b = _get_mif(node, "b")

    a_one = _get_mif(a, "one")
    b_one = _get_mif(b, "one")
    a_two = _get_mif(a, "two")
    b_two = _get_mif(b, "two")

    assert a_one.is_connected_to(b_one)
    assert a_two.is_connected_to(b_two)
    assert not any(a_one.is_connected_to(other) for other in [a_two, b_two])
    assert not any(a_two.is_connected_to(other) for other in [a_one, b_one])


def test_shim_power(bob: Bob):
    from atopile.attributes import Power

    a = Power()
    b = F.ElectricPower()

    bob._connect(a, b, None)

    assert a.lv.is_connected_to(b.lv)
    assert a.hv.is_connected_to(b.hv)
    assert not a.lv.is_connected_to(b.hv)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class PowerSwitchMOSFET(F.PowerSwitch):
    """
    Power switch using a MOSFET

    This power switch uses an NMOS when lowside, and a PMOS when highside.
    """

    def __init__(self, lowside: bool, normally_closed: bool) -> None:
        super().__init__(normally_closed=normally_closed)

        self._lowside = lowside

    # components

    mosfet: F.MOSFET

    def __preinit__(self):
        self.mosfet.channel_type.constrain_subset(
            F.MOSFET.ChannelType.N_CHANNEL
            if self._lowside
            else F.MOSFET.ChannelType.P_CHANNEL
        )
        self.mosfet.saturation_type.constrain_subset(
            F.MOSFET.SaturationType.ENHANCEMENT
        )

        # pull gate
        # lowside     normally_closed   pull up
        # True        True              True
        # True        False             False
        # False       True              False
        # False       False             True
        self.logic_in.pulled.pull(self._lowside == self._normally_closed, owner=self)

        # connect gate to logic
        self.logic_in.line.connect(self.mosfet.gate)

        # passthrough non-switched side, bridge switched side
        if self._lowside:
            self.power_in.hv.connect(self.switched_power_out.hv)
            self.power_in.lv.connect_via(self.mosfet, self.switched_power_out.lv)
        else:
            self.power_in.lv.connect(self.switched_power_out.lv)
            self.power_in.hv.connect_via(self.mosfet, self.switched_power_out.hv)

        # TODO do more with logic
        #   e.g check reference being same as power



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L


class SWDConnector(Module):
    swd: F.SWD
    gnd_detect: F.ElectricLogic
    reference: F.ElectricPower

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.J
    )

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://developer.arm.com/documentation/101636/0100/Debug-and-Trace/JTAG-SWD-Interface"  # noqa E501
    )

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        pass



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import logging
from enum import Enum

import pytest

from faebryk.core.parameter import ParameterOperableHasNoLiteral
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


def test_repr_chain_basic():
    import faebryk.library._F as F

    class TestModule(L.Module):
        param1 = L.p_field(units=P.V)
        param2 = L.p_field(units=P.A)
        param3 = L.p_field(units=P.V)

        @L.rt_field
        def simple_value_representation(self):
            S = F.has_simple_value_representation_based_on_params_chain.Spec
            return F.has_simple_value_representation_based_on_params_chain(
                S(param=self.param1, tolerance=True),
                S(param=self.param2, suffix="P2"),
                S(param=self.param3, tolerance=True, suffix="P3"),
                prefix="TM",
            )

    m = TestModule()
    m.param1.alias_is(L.Range(10 * P.V, 20 * P.V))
    m.param2.alias_is(5 * P.A)
    m.param3.alias_is(10 * P.V)

    val = m.get_trait(F.has_simple_value_representation).get_value()
    assert val == "TM 15V ±33% 5A P2 10V P3"


def test_repr_chain_non_number():
    import faebryk.library._F as F

    class TestEnum(Enum):
        A = "AS"
        B = "BS"

    class TestModule(L.Module):
        param1 = L.p_field(domain=L.Domains.ENUM(TestEnum))
        param2 = L.p_field(domain=L.Domains.BOOL())

        @L.rt_field
        def simple_value_representation(self):
            S = F.has_simple_value_representation_based_on_params_chain.Spec
            return F.has_simple_value_representation_based_on_params_chain(
                S(param=self.param1),
                S(param=self.param2, prefix="P2:"),
            )

    m = TestModule()
    m.param1.alias_is(TestEnum.A)
    m.param2.alias_is(True)

    val = m.get_trait(F.has_simple_value_representation).get_value()
    assert val == "AS P2: true"


def test_repr_chain_no_literal():
    import faebryk.library._F as F

    class TestModule(L.Module):
        param1 = L.p_field(units=P.V)
        param2 = L.p_field(units=P.A)
        param3 = L.p_field(units=P.V)

        @L.rt_field
        def simple_value_representation(self):
            S = F.has_simple_value_representation_based_on_params_chain.Spec
            return F.has_simple_value_representation_based_on_params_chain(
                S(param=self.param1, default=None),
                S(param=self.param2),
                S(param=self.param3, default="P3: MISSING"),
            )

    m = TestModule()

    pytest.raises(
        ParameterOperableHasNoLiteral,
        m.get_trait(F.has_simple_value_representation).get_value,
    )

    m.param1.alias_is(10 * P.V)
    val = m.get_trait(F.has_simple_value_representation).get_value()
    assert val == "10V P3: MISSING"



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import itertools
import logging

import pytest

import faebryk.library._F as F
from faebryk.exporters.netlist.graph import attach_nets
from faebryk.exporters.netlist.kicad.netlist_kicad import (
    attach_kicad_info,
    faebryk_netlist_to_kicad,
)
from faebryk.exporters.netlist.netlist import FBRKNetlist, make_fbrk_netlist_from_graph
from faebryk.libs.app.designators import (
    attach_random_designators,
)
from faebryk.libs.kicad.fileformats import C_fields, C_kicad_netlist_file
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


# Netlists --------------------------------------------------------------------
@pytest.fixture
def netlist_graph():
    resistor1 = F.Resistor().builder(lambda r: r.resistance.alias_is(100 * P.ohm))
    resistor2 = F.Resistor().builder(lambda r: r.resistance.alias_is(200 * P.ohm))
    power = F.ElectricPower()

    # net labels
    vcc = F.Net.with_name("+3V3")
    gnd = F.Net.with_name("GND")
    power.hv.connect(vcc.part_of)
    power.lv.connect(gnd.part_of)

    # connect
    resistor1.unnamed[0].connect(power.hv)
    resistor1.unnamed[1].connect(power.lv)
    resistor2.unnamed[0].connect(resistor1.unnamed[0])
    resistor2.unnamed[1].connect(resistor1.unnamed[1])

    # attach footprint & designator
    for i, r in enumerate([resistor1, resistor2]):
        r.get_trait(F.can_attach_to_footprint).attach(
            F.SMDTwoPin(F.SMDTwoPin.Type._0805)
        )
        r.add(
            F.has_designator(
                resistor1.get_trait(F.has_designator_prefix).get_prefix() + str(i + 1)
            )
        )

    # make netlist
    G = resistor1.get_graph
    attach_random_designators(G())
    attach_nets(G())
    attach_kicad_info(G())
    return make_fbrk_netlist_from_graph(G())


@pytest.fixture
def netlist_t2():
    # t2_netlist = [(properties, vertices=[comp=(name, value, properties), pin)])]
    resistor1 = FBRKNetlist.Component(
        name="R1",
        value="100Ω",
        properties={
            "footprint": "Resistor_SMD:R_0805_2012Metric",
        },
    )

    resistor2 = FBRKNetlist.Component(
        name="R2",
        value="200Ω",
        properties={
            "footprint": "Resistor_SMD:R_0805_2012Metric",
        },
    )

    return FBRKNetlist(
        nets=[
            FBRKNetlist.Net(
                properties={
                    "name": "GND",
                },
                vertices=[
                    FBRKNetlist.Net.Vertex(
                        component=resistor1,
                        pin="2",
                    ),
                    FBRKNetlist.Net.Vertex(
                        component=resistor2,
                        pin="2",
                    ),
                ],
            ),
            FBRKNetlist.Net(
                properties={
                    "name": "+3V3",
                },
                vertices=[
                    FBRKNetlist.Net.Vertex(
                        component=resistor1,
                        pin="1",
                    ),
                    FBRKNetlist.Net.Vertex(
                        component=resistor2,
                        pin="1",
                    ),
                ],
            ),
        ],
        comps=[resistor1, resistor2],
    )


@pytest.fixture
def netlist_manu():
    # Footprint pins are just referenced by number through netlist of symbol
    # Careful comps need distinct timestamps
    tstamp = itertools.count(1)

    N = C_kicad_netlist_file
    return C_kicad_netlist_file(
        N.C_netlist(
            version="E",
            components=N.C_netlist.C_components(
                comps=[
                    N.C_netlist.C_components.C_component(
                        ref="R1",
                        value="100Ω",
                        footprint="Resistor_SMD:R_0805_2012Metric",
                        tstamps=str(next(tstamp)),
                        fields=C_fields(fields={}),
                        propertys={},
                    ),
                    N.C_netlist.C_components.C_component(
                        ref="R2",
                        value="200Ω",
                        footprint="Resistor_SMD:R_0805_2012Metric",
                        tstamps=str(next(tstamp)),
                        fields=C_fields(fields={}),
                        propertys={},
                    ),
                ]
            ),
            nets=N.C_netlist.C_nets(
                nets=[
                    N.C_netlist.C_nets.C_net(
                        code=1,
                        name="+3V3",
                        nodes=[
                            N.C_netlist.C_nets.C_net.C_node(
                                ref="R1",
                                pin="1",
                            ),
                            N.C_netlist.C_nets.C_net.C_node(
                                ref="R2",
                                pin="1",
                            ),
                        ],
                    ),
                    N.C_netlist.C_nets.C_net(
                        code=2,
                        name="GND",
                        nodes=[
                            N.C_netlist.C_nets.C_net.C_node(
                                ref="R1",
                                pin="2",
                            ),
                            N.C_netlist.C_nets.C_net.C_node(
                                ref="R2",
                                pin="2",
                            ),
                        ],
                    ),
                ]
            ),
        ),
    )


def test_netlist_t2(netlist_t2, netlist_manu):
    netlist_t2 = faebryk_netlist_to_kicad(netlist_t2)
    assert netlist_t2.dumps() == netlist_manu.dumps()


def test_netlist_graph(netlist_graph, netlist_t2):
    for comp in netlist_graph.comps:
        assert isinstance(comp.properties, dict)
        comp.properties.pop("atopile_address", None)

    netlist_graph = faebryk_netlist_to_kicad(netlist_graph)
    netlist_t2 = faebryk_netlist_to_kicad(netlist_t2)
    assert netlist_graph.dumps() == netlist_t2.dumps()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from abc import abstractmethod
from enum import IntEnum

from faebryk.core.module import Module


class has_pcb_position(Module.TraitT):
    class layer_type(IntEnum):
        NONE = 0
        TOP_LAYER = -1
        BOTTOM_LAYER = 1

    Point = tuple[float, float, float, layer_type]

    @abstractmethod
    def get_position(self) -> Point: ...



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from abc import abstractmethod
from typing import TYPE_CHECKING

import faebryk.library._F as F

if TYPE_CHECKING:
    from faebryk.library.Pad import Pad


class has_kicad_footprint(F.Footprint.TraitT):
    @abstractmethod
    def get_kicad_footprint(self) -> str: ...

    @abstractmethod
    def get_pin_names(self) -> dict["Pad", str]: ...

    def get_kicad_footprint_name(self) -> str:
        return self.get_kicad_footprint().split(":")[-1]



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from faebryk.libs.library import L


class is_app_root(L.Module.TraitT.decless()):
    """
    Indicates that the module is the root of an application.
    """

    def on_obj_set(self):
        obj = self.get_obj(L.Node)
        obj.no_include_parents_in_full_name = True



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.util import assert_once, times

logger = logging.getLogger(__name__)


class CH342F(F.CH342):
    """
    USB to double Base UART converter

    QFN-24-EP(4x4)
    """

    class UARTWrapper(Module):
        uart: F.UART
        tnow: F.ElectricLogic
        ...

        @assert_once
        def enable_tnow_mode(self, owner: Module):
            """
            Set TNOW mode for specified UART for use with RS485 tranceivers.
            The TNOW pin can be connected to the tx_enable and rx_enable
            pins of the RS485 tranceiver for automatic half-duplex control.
            """
            self.uart.dtr.set_weak(on=False, owner=owner)
            self.uart.dtr.connect(self.tnow)

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    vbus_detect: F.ElectricLogic

    reset: F.ElectricLogic
    active: F.ElectricLogic

    uart = times(2, UARTWrapper)

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    @L.rt_field
    def can_attach_to_footprint(self):
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": self.uart[0].uart.ri.line,
                "2": self.usb.usb_if.buspower.lv,
                "3": self.usb.usb_if.d.p.line,
                "4": self.usb.usb_if.d.n.line,
                "5": self.power_io.hv,
                "6": self.power_3v.hv,
                "7": self.integrated_regulator.power_in.hv,
                "8": self.usb.usb_if.buspower.hv,
                "9": self.reset.line,
                "10": self.uart[1].uart.cts.line,
                "11": self.uart[1].uart.rts.line,
                "12": self.uart[1].uart.base_uart.rx.line,
                "13": self.uart[1].uart.base_uart.tx.line,
                "14": self.uart[1].uart.dsr.line,
                "15": self.uart[1].uart.dtr.line,
                "16": self.uart[1].uart.dcd.line,
                "17": self.uart[1].uart.ri.line,
                "18": self.uart[0].uart.cts.line,
                "19": self.uart[0].uart.rts.line,
                "20": self.uart[0].uart.base_uart.rx.line,
                "21": self.uart[0].uart.base_uart.tx.line,
                "22": self.uart[0].uart.dsr.line,
                "23": self.uart[0].uart.dtr.line,
                "24": self.uart[0].uart.dcd.line,
                "25": self.usb.usb_if.buspower.lv,
            }
        )

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.uart[0].uart.cts.line: ["CTS0"],
                self.uart[1].uart.cts.line: ["CTS1"],
                self.uart[0].uart.dcd.line: ["DCD0"],
                self.uart[1].uart.dcd.line: ["DCD1"],
                self.uart[0].uart.dsr.line: ["DSR0"],
                self.uart[1].uart.dsr.line: ["DSR1"],
                self.uart[0].uart.dtr.line: ["DTR0"],
                self.uart[1].uart.dtr.line: ["DTR1"],
                # self.power_3v.lv: ["EP"],
                self.power_3v.lv: ["GND"],
                self.uart[0].uart.ri.line: ["RI0"],
                self.uart[1].uart.ri.line: ["RI1"],
                self.reset.line: ["RST"],
                self.uart[0].uart.rts.line: ["RTS0"],
                self.uart[1].uart.rts.line: ["RTS1"],
                self.uart[0].uart.base_uart.rx.line: ["RXD0"],
                self.uart[1].uart.base_uart.rx.line: ["RXD1"],
                self.uart[0].uart.base_uart.tx.line: ["TXD0"],
                self.uart[1].uart.base_uart.tx.line: ["TXD1"],
                self.usb.usb_if.d.p.line: ["UD+"],
                self.usb.usb_if.d.n.line: ["UD-"],
                self.power_3v.hv: ["V3"],
                self.vbus_detect.line: ["VBUS"],
                self.integrated_regulator.power_in.hv: ["VDD5"],
                self.power_io.hv: ["VIO"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    descriptive_properties = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "WCH(Jiangsu Qin Heng)",
            DescriptiveProperties.partno: "CH342F",
        }
    )

    def __preinit__(self) -> None:
        # ----------------------------------------
        #                aliasess
        # ----------------------------------------

        # ----------------------------------------
        #            parametrization
        # ----------------------------------------

        # ----------------------------------------
        #              connections
        # ----------------------------------------
        pass
        # TODO: specialize base uarts from CH342 base class
        # uarts = times(2, F.UART)
        # for uart, uart_base in zip(uarts, self.uart_base):
        #    uart_base.specialize(uart)
        #    self.add(uart)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from pathlib import Path

import pytest

import faebryk.library._F as F  # noqa: F401  # This is required to prevent a circular import
from faebryk.libs.kicad.fileformats import (
    C_effects,
    C_footprint,
    C_kicad_footprint_file,
    C_kicad_fp_lib_table_file,
    C_kicad_netlist_file,
    C_kicad_pcb_file,
    C_kicad_project_file,
)
from faebryk.libs.kicad.fileformats_sch import C_kicad_sch_file, C_kicad_sym_file
from faebryk.libs.kicad.fileformats_version import kicad_footprint_file
from faebryk.libs.sexp.dataclass_sexp import JSON_File, SEXP_File
from faebryk.libs.test.fileformats import (
    _FP_DIR,
    _FPLIB_DIR,  # noqa: F401
    _NETLIST_DIR,  # noqa: F401
    _PCB_DIR,  # noqa: F401
    _PRJ_DIR,  # noqa: F401
    _SCH_DIR,  # noqa: F401
    _SYM_DIR,  # noqa: F401
    _VERSION_DIR,  # noqa: F401
    DEFAULT_VERSION,  # noqa: F401
    FPFILE,
    FPLIBFILE,
    NETFILE,
    PCBFILE,
    PRJFILE,
    SCHFILE,
    SYMFILE,
)
from faebryk.libs.util import ConfigFlag, find, not_none

logger = logging.getLogger(__name__)

DUMP = ConfigFlag("DUMP", descr="dump load->save into /tmp")


def test_parser():
    pcb = C_kicad_pcb_file.loads(PCBFILE)
    fp = C_kicad_footprint_file.loads(FPFILE)
    netlist = C_kicad_netlist_file.loads(NETFILE)
    pro = C_kicad_project_file.loads(PRJFILE)
    sch = C_kicad_sch_file.loads(SCHFILE)
    sym = C_kicad_sym_file.loads(SYMFILE)

    assert [f.name for f in pcb.kicad_pcb.footprints] == [
        "logos:faebryk_logo",
        "lcsc:LED0603-RD-YELLOW",
        "lcsc:R0402",
        "lcsc:BAT-TH_BS-02-A1AJ010",
    ]

    assert not pcb.kicad_pcb.setup.pcbplotparams.usegerberextensions

    padtype = pcb.C_kicad_pcb.C_pcb_footprint.C_pad.E_type
    assert [(p.name, p.type) for p in fp.footprint.pads] == [
        ("", padtype.smd),
        ("", padtype.smd),
        ("1", padtype.smd),
        ("2", padtype.smd),
    ]

    assert [(c.ref, c.value) for c in netlist.export.components.comps][:10] == [
        ("C1", "10uF"),
        ("C2", "10uF"),
        ("C3", "10uF"),
        ("C4", "10uF"),
        ("C5", "22uF"),
        ("C6", "100nF"),
        ("C7", "100nF"),
        ("C8", "10uF"),
        ("C9", "100nF"),
        ("C10", "100nF"),
    ]

    # Var args parser
    effects = (
        find(pcb.kicad_pcb.footprints, lambda f: f.name == "logos:faebryk_logo")
        .propertys["Footprint"]
        .effects
    )
    assert effects.justifys[0].justifys == [
        C_effects.C_justify.E_justify.mirror,
        C_effects.C_justify.E_justify.right,
    ]
    assert effects.justifys[1].justifys == [
        C_effects.C_justify.E_justify.bottom,
    ]

    assert effects.get_justifys() == [
        C_effects.C_justify.E_justify.mirror,
        C_effects.C_justify.E_justify.right,
        C_effects.C_justify.E_justify.bottom,
    ]

    assert pro.pcbnew.last_paths.netlist == "../../faebryk/faebryk.net"

    assert (
        sch.kicad_sch.lib_symbols.symbols["Amplifier_Audio:LM4990ITL"]
        .propertys["Datasheet"]
        .value
        == "http://www.ti.com/lit/ds/symlink/lm4990.pdf"
    )

    assert sch.kicad_sch.lib_symbols.symbols["power:GND"].power is not None
    assert sch.kicad_sch.lib_symbols.symbols["Device:R"].power is None

    assert (
        sym.kicad_symbol_lib.symbols["AudioJack-CUI-SJ-3523-SMT"].name
        == "AudioJack-CUI-SJ-3523-SMT"
    )


def test_write():
    pcb = C_kicad_pcb_file.loads(PCBFILE)

    def _d1(pcb: C_kicad_pcb_file):
        return find(
            pcb.kicad_pcb.footprints,
            lambda f: f.propertys["Reference"].value == "D1",
        )

    led_p = _d1(pcb).propertys["Value"]
    assert led_p.value == "LED"
    led_p.value = "LED2"

    pcb_reload = C_kicad_pcb_file.loads(pcb.dumps())
    assert _d1(pcb_reload).propertys["Value"].value == "LED2"


def test_empty_enum_positional():
    pcb = C_kicad_pcb_file.loads(PCBFILE)

    def _b1_p1(pcb: C_kicad_pcb_file):
        return find(
            find(
                pcb.kicad_pcb.footprints,
                lambda f: f.propertys["Reference"].value == "B1",
            ).pads,
            lambda p: p.name == "1",
        )

    _b1_p1(pcb).drill = C_footprint.C_pad.C_drill(
        C_footprint.C_pad.C_drill.E_shape.stadium, 0.5, 0.4
    )

    def _effects(pcb: C_kicad_pcb_file):
        return (
            find(pcb.kicad_pcb.footprints, lambda f: f.name == "logos:faebryk_logo")
            .propertys["Datasheet"]
            .effects
        )

    _effects(pcb).justifys.append(
        C_effects.C_justify([C_effects.C_justify.E_justify.center_horizontal])
    )
    _effects(pcb).justifys.append(
        C_effects.C_justify([C_effects.C_justify.E_justify.top])
    )

    pcb_reload = C_kicad_pcb_file.loads(pcb.dumps())

    assert (
        not_none(_b1_p1(pcb_reload).drill).shape
        == C_footprint.C_pad.C_drill.E_shape.stadium
    )

    # empty center string ignored
    assert _effects(pcb).get_justifys() == [
        C_effects.C_justify.E_justify.center_horizontal,
        C_effects.C_justify.E_justify.top,
    ]


@pytest.mark.parametrize(
    ("parser", "path"),
    [
        (C_kicad_pcb_file, PCBFILE),
        (C_kicad_footprint_file, FPFILE),
        (C_kicad_netlist_file, NETFILE),
        (C_kicad_project_file, PRJFILE),
        (C_kicad_fp_lib_table_file, FPLIBFILE),
        (C_kicad_sch_file, SCHFILE),
        (C_kicad_sym_file, SYMFILE),
    ],
)
def test_dump_load_equality(parser: type[SEXP_File | JSON_File], path: Path):
    loaded = parser.loads(path)
    dump = loaded.dumps(Path("/tmp") / path.name if DUMP else None)
    loaded_dump = parser.loads(dump)
    dump2 = loaded_dump.dumps()
    assert dump == dump2


@pytest.mark.parametrize(
    "fp_path", _FP_DIR(5).glob("*.kicad_mod"), ids=lambda p: p.stem
)
def test_v5_fp_convert(fp_path: Path):
    fp = kicad_footprint_file(fp_path)
    assert fp.footprint.name.split(":")[-1] == fp_path.stem


@pytest.mark.parametrize(
    "fp_path", _FP_DIR(6).glob("*.kicad_mod"), ids=lambda p: p.stem
)
def test_v6_fp_convert(fp_path: Path):
    fp = kicad_footprint_file(fp_path)
    assert fp.footprint.name.split(":")[-1] == fp_path.stem



import contextlib
import json
import logging
from pathlib import Path
from typing import Any, Iterable, Optional

import pcbnew  # type: ignore

log = logging.getLogger(__name__)


@contextlib.contextmanager
def log_exceptions():
    try:
        yield
    except Exception as e:
        log.exception(e)
        raise


def get_prj_dir(path: Path) -> Path:
    """Return the atopile project directory."""
    path = Path(path)
    if (path / "ato.yaml").exists():
        return path
    for p in path.parents:
        if (p / "ato.yaml").exists():
            return p
    raise FileNotFoundError("ato.yaml not found in any parent directory")


def get_board_artifact_manifest(board_path: Path) -> dict:
    """Return a dict of the artifact manifest related to this board."""
    board_path = Path(board_path)
    manifest_path = get_prj_dir(board_path) / "build" / "manifest.json"
    with manifest_path.open("r") as f:
        manifest = json.load(f)
    return manifest.get("by-layout", {}).get(str(board_path), {})


def groups_by_name(board: pcbnew.BOARD) -> dict[str, pcbnew.PCB_GROUP]:
    """Return a dict of groups by name."""
    return {g.GetName(): g for g in board.Groups()}


def get_footprint_uuid(fp: pcbnew.FOOTPRINT) -> str:
    """Return the UUID of a footprint."""
    path = fp.GetPath().AsString()
    return path.split("/")[-1]


def get_footprint_addr(fp: pcbnew.FOOTPRINT, uuid_map: dict[str, str]) -> Optional[str]:
    """Return the property "atopile_address" of a footprint."""
    field: pcbnew.PCB_FIELD = fp.GetFieldByName("atopile_address")
    if field:
        return field.GetText()

    # This is a backup to support old layouts which
    # haven't yet got addresses in their properties
    uuid = get_footprint_uuid(fp)
    if uuid in uuid_map:
        return uuid_map[uuid]

    return None


def footprints_by_addr(
    board: pcbnew.BOARD, uuid_map: dict[str, str]
) -> dict[str, pcbnew.FOOTPRINT]:
    """Return a dict of footprints by "atopile_address"."""
    return {
        addr: fp
        for fp in board.GetFootprints()
        if (addr := get_footprint_addr(fp, uuid_map))
    }


def get_layout_map(board_path: Path) -> dict[str, Any]:
    """Return the layout map for the board."""
    board_path = Path(board_path)
    manifest = get_board_artifact_manifest(board_path)
    with Path(manifest["layouts"]).open("r", encoding="utf-8") as f:
        return json.load(f)


def flip_dict(d: dict) -> dict:
    """Return a dict with keys and values swapped."""
    return {v: k for k, v in d.items()}


def sync_track(
    source_board: pcbnew.BOARD,
    track: pcbnew.PCB_TRACK,
    target_board: pcbnew.BOARD,
    net_map: dict[str, str],
) -> pcbnew.PCB_TRACK:
    """Sync a track to the target board and update its net from net_map."""
    new_track: pcbnew.PCB_TRACK = track.Duplicate().Cast()
    new_track.SetParent(target_board)
    new_track.SetStart(track.GetStart())
    new_track.SetEnd(track.GetEnd())
    new_track.SetLayer(track.GetLayer())

    # Update the net using net_map
    source_net = track.GetNetname()
    if source_net in net_map:
        target_net_name = net_map[source_net]
        # Find the actual NETINFO_ITEM for the target net
        target_net_info = target_board.FindNet(target_net_name)
        if target_net_info:
            new_track.SetNet(target_net_info)
        else:
            log.warning(f"Could not find net '{target_net_name}' in target board")
            new_track.SetNetCode(0)  # No net
    else:
        new_track.SetNetCode(0)  # No net

    target_board.Add(new_track)
    source_board.Remove(new_track)
    return new_track


def generate_net_map(
    source_board: pcbnew.BOARD,
    target_board: pcbnew.BOARD,
    addr_map: dict[str, str],
    uuid_map: dict[str, str],
):
    """Generates a mapping from source net names to target net names based on matching
    addresses.

    For each component in the source board, if its address exists in addr_map, the
    target footprint is looked up. For each pad index in the source footprint, a mapping
    is added from source pad net (e.g., 'VCC') to target pad net (e.g., 'VCC_3V3').
    """
    net_map: dict[str, str] = {}
    # Track how many times we've seen each mapping for confidence
    mapping_counts: dict[str, dict[str, int]] = {}

    # Obtain a dictionary of target footprints by their address
    target_fps = footprints_by_addr(target_board, uuid_map)

    # Iterate through each footprint in the source board
    for src_fp in source_board.GetFootprints():
        # Get the source footprint address
        src_addr = get_footprint_addr(src_fp, uuid_map)
        if not src_addr:
            continue

        # Check if there is a mapping for this source address
        if src_addr not in addr_map:
            continue
        target_addr = addr_map[src_addr]

        # Look up the corresponding target footprint
        target_fp = target_fps.get(target_addr)
        if not target_fp:
            continue

        # For each pad in the source footprint, map its net to the pad in the target
        source_pads = list(src_fp.Pads())
        target_pads = list(target_fp.Pads())

        # Match pads by pad number, handling duplicate pad numbers by comparing size
        for source_pad in source_pads:
            pad_number = source_pad.GetNumber()
            source_net = source_pad.GetNetname()

            # Find matching target pads with the same number
            matching_target_pads = [
                p for p in target_pads if p.GetNumber() == pad_number
            ]

            target_net = None

            if len(matching_target_pads) == 1:
                # Exact match by number
                target_pad = matching_target_pads[0]
                target_net = target_pad.GetNetname()
            elif len(matching_target_pads) > 1:
                # Multiple pads with same number, find best match by size and shape
                best_match = None
                min_diff = float("inf")

                for candidate in matching_target_pads:
                    # Calculate difference in size and shape
                    size_diff = abs(
                        candidate.GetSize().x - source_pad.GetSize().x
                    ) + abs(candidate.GetSize().y - source_pad.GetSize().y)
                    shape_diff = (
                        0 if candidate.GetShape() == source_pad.GetShape() else 10
                    )
                    total_diff = size_diff + shape_diff

                    if total_diff < min_diff:
                        min_diff = total_diff
                        best_match = candidate

                if best_match:
                    target_net = best_match.GetNetname()

            # If we found a target net, update the mapping
            if target_net and source_net:
                # Initialize count tracking for this source net if we haven't seen it
                if source_net not in mapping_counts:
                    mapping_counts[source_net] = {}

                # Increment the count for this specific mapping
                current_count = mapping_counts[source_net].get(target_net, 0)
                mapping_counts[source_net][target_net] = current_count + 1

                # If we've seen this source net before, verify the mapping
                if source_net in net_map:
                    # If the existing mapping doesn't match, log a warning
                    if net_map[source_net] != target_net:
                        log.warning(
                            f"Conflicting net mapping for '{source_net}': "
                            f"'{net_map[source_net]}' vs '{target_net}'. "
                            f"Using most frequently seen mapping."
                        )

                        # Use the most frequently seen mapping
                        most_frequent_target = max(
                            mapping_counts[source_net].items(), key=lambda x: x[1]
                        )[0]
                        net_map[source_net] = most_frequent_target
                else:
                    # First time seeing this source net
                    net_map[source_net] = target_net

    # Log mapping confidence statistics
    for source_net, counts in mapping_counts.items():
        if len(counts) > 1:
            log.info(
                f"Multiple mappings found for source net '{source_net}': {counts}"
                f" - Using: '{net_map[source_net]}'"
            )
        else:
            # Single mapping with count
            target_net, count = next(iter(counts.items()))
            log.debug(
                f"Net mapping '{source_net}' -> '{target_net}' "
                f"confirmed {count} times"
            )

    return net_map


# also pull in any silkscreen items
def sync_drawing(
    source_board: pcbnew.BOARD, drawing: pcbnew.DRAWINGS, target_board: pcbnew.BOARD
) -> pcbnew.DRAWINGS:
    """Sync a drawing to the target board."""
    new_drawing: pcbnew.DRAWINGS = drawing.Duplicate().Cast()
    new_drawing.SetParent(target_board)
    new_drawing.SetLayer(drawing.GetLayer())
    target_board.Add(new_drawing)
    source_board.Remove(new_drawing)
    return new_drawing


def update_zone_net(
    source_zone: pcbnew.ZONE,
    target_zone: pcbnew.ZONE,
    net_map: dict[str, str],
):
    """Updates the target zone's net using net_map.

    If the source zone's netname exists in net_map, the target zone's net is set
    to the corresponding value. Otherwise, the net is set to no net (net code 0).
    """
    source_netname = source_zone.GetNetname()
    log.info(
        f"update_zone_net: source_zone net = {source_netname} "
        f"for target_zone {target_zone}"
    )
    if source_netname in net_map:
        target_net_name = net_map[source_netname]
        # We need to get the actual NETINFO_ITEM from the target board
        target_board = target_zone.GetBoard()
        target_net_info = target_board.FindNet(target_net_name)
        if target_net_info:
            target_zone.SetNet(target_net_info)
        else:
            log.warning(
                f"Could not find net '{target_net_name}' in target board for zone"
            )
            target_zone.SetNetCode(0)
    else:
        target_zone.SetNetCode(0)


def sync_zone(zone: pcbnew.ZONE, target: pcbnew.BOARD) -> pcbnew.ZONE:
    """Sync a zone to the target board."""
    new_zone: pcbnew.ZONE = zone.Duplicate().Cast()
    new_zone.SetParent(target)
    layer = zone.GetLayer()
    new_zone.SetNet(target.FindNet(zone.GetNetname()))
    if layer != -1:
        new_zone.SetLayer(zone.GetLayer())
    target.Add(new_zone)
    return new_zone


def sync_footprints(
    source: pcbnew.BOARD,
    target: pcbnew.BOARD,
    addr_map: dict[str, str],
    uuid_map: dict[str, str],
) -> list[str]:
    """Update the target board with the layout from the source board."""
    # Update the footprint position, orientation, and side to match the source
    source_addrs = footprints_by_addr(source, uuid_map)
    target_addrs = footprints_by_addr(target, uuid_map)
    missing_addrs = []

    for s_addr, t_addr in addr_map.items():
        try:
            target_fp = target_addrs[t_addr]
            source_fp = source_addrs[s_addr]
        except KeyError:
            missing_addrs.append(s_addr)
            continue

        target_fp.SetPosition(source_fp.GetPosition())
        target_fp.SetOrientation(source_fp.GetOrientation())
        target_fp.SetLayer(source_fp.GetLayer())
        target_fp.SetLayerSet(source_fp.GetLayerSet())
        # Ref Designators
        target_fp.Reference().SetAttributes(source_fp.Reference().GetAttributes())
        target_ref: pcbnew.FP_TEXT = target_fp.Reference()
        source_ref: pcbnew.FP_TEXT = source_fp.Reference()
        target_ref.SetPosition(source_ref.GetPosition())

        # Pads
        source_pads = list(source_fp.Pads())
        target_pads = list(target_fp.Pads())

        # First try to match by pad number
        for target_pad in target_pads:
            pad_number = target_pad.GetNumber()
            matching_source_pads = [
                p for p in source_pads if p.GetNumber() == pad_number
            ]

            if len(matching_source_pads) == 1:
                # Exact match by number
                source_pad = matching_source_pads[0]
            elif len(matching_source_pads) > 1:
                # Multiple pads with same number, find best match by size and shape
                best_match = None
                min_diff = float("inf")

                for candidate in matching_source_pads:
                    # Calculate difference in size and shape
                    size_diff = abs(
                        candidate.GetSize().x - target_pad.GetSize().x
                    ) + abs(candidate.GetSize().y - target_pad.GetSize().y)
                    shape_diff = (
                        0 if candidate.GetShape() == target_pad.GetShape() else 10
                    )
                    total_diff = size_diff + shape_diff

                    if total_diff < min_diff:
                        min_diff = total_diff
                        best_match = candidate

                source_pad = best_match
            else:
                # No matching pad by number, try to find closest by size and position
                best_match = None
                min_diff = float("inf")

                for candidate in source_pads:
                    # Calculate difference in size
                    size_diff = abs(
                        candidate.GetSize().x - target_pad.GetSize().x
                    ) + abs(candidate.GetSize().y - target_pad.GetSize().y)

                    # Only consider pads with similar size (threshold can be adjusted)
                    if size_diff > 1.0:  # mm
                        continue

                    # If similar size, consider shape and layer
                    shape_diff = (
                        0 if candidate.GetShape() == target_pad.GetShape() else 10
                    )
                    layer_diff = (
                        0 if candidate.GetLayer() == target_pad.GetLayer() else 5
                    )

                    total_diff = size_diff + shape_diff + layer_diff

                    if total_diff < min_diff:
                        min_diff = total_diff
                        best_match = candidate

                if best_match is not None:
                    source_pad = best_match
                    log.warning(
                        f"Used fallback matching for pad {pad_number} in {t_addr}"
                    )
                else:
                    log.warning(
                        f"No suitable pad match found for {pad_number} in {s_addr}"
                    )
                    continue

            # Make sure we found a valid source pad before proceeding
            if source_pad is None:
                log.warning(
                    f"Failed to find suitable source pad match for target"
                    f" pad {pad_number} in {s_addr}"
                )
                continue

            # Apply the source pad's position and orientation to the target pad
            target_pad.SetPosition(source_pad.GetPosition())
            target_pad.SetOrientation(source_pad.GetOrientation())
            target_pad.SetLayer(source_pad.GetLayer())
            target_pad.SetLayerSet(source_pad.GetLayerSet())

    return missing_addrs


def find_anchor_footprint(fps: Iterable[pcbnew.FOOTPRINT]) -> pcbnew.FOOTPRINT:
    """
    Return anchor footprint with largest pin count in board or group: tiebreaker size
    """
    # fps = layout.GetFootprints()
    # fps = [item for item in layout.GetItems() if isinstance(item, pcbnew.FOOTPRINT)]
    max_padcount = 0
    max_area = 0
    anchor_fp = None
    for fp in fps:
        if fp.GetPadCount() > max_padcount or (
            fp.GetPadCount() == max_padcount and fp.GetArea() > max_area
        ):
            anchor_fp = fp
            max_padcount = fp.GetPadCount()
            max_area = fp.GetArea()
    return anchor_fp


def get_group_footprints(group: pcbnew.PCB_GROUP) -> list[pcbnew.FOOTPRINT]:
    """Return a list of footprints in a group."""
    return [item for item in group.GetItems() if isinstance(item, pcbnew.FOOTPRINT)]


def calculate_translation(
    source_fps: Iterable[pcbnew.FOOTPRINT], target_fps: Iterable[pcbnew.FOOTPRINT]
) -> pcbnew.VECTOR2I:
    """Calculate the translation vector between two groups of footprints."""
    source_anchor_fp = find_anchor_footprint(source_fps)
    source_offset = source_anchor_fp.GetPosition()
    target_anchor_fp = find_anchor_footprint(target_fps)
    target_offset = target_anchor_fp.GetPosition()
    total_offset = target_offset - source_offset
    return total_offset



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import ast
from pathlib import Path

import pytest

from faebryk.libs.util import repo_root as _repo_root


def _extract_classes_from_file(filepath: Path):
    with open(filepath, "r") as file:
        module = ast.parse(file.read())
    classes = [node for node in module.body if isinstance(node, ast.ClassDef)]
    return classes


@pytest.mark.parametrize(
    "py_file",
    [
        p
        for p in (_repo_root() / "src" / "faebryk" / "library").glob("**/*.py")
        if p.is_file()
    ],
    ids=lambda p: p.stem,
)
def test_class_name(py_file: Path):
    """Test that class names match their file names in the library directory."""
    classes = _extract_classes_from_file(py_file)
    for cls in classes:
        if cls.name.startswith("_"):
            continue
        assert cls.name == py_file.stem, f"Class name mismatch in {py_file}"



import contextlib
import sys

import rich

from atopile import telemetry
from faebryk.libs.logging import FLOG_FMT


def _handle_exception(exc_type, exc_value, exc_traceback):
    # avoid exceptions raised during import
    from atopile.cli.logging import logger
    from atopile.errors import _BaseBaseUserException

    # delayed import to improve startup time
    from faebryk.libs.util import in_debug_session

    if issubclass(exc_type, _BaseBaseUserException):
        # If we're in a debug session, we want to see the
        # unadulterated exception. We do this pre-logging because
        # we don't want the logging to potentially obstruct the debugger.
        if in_debug_session():
            import debugpy

            debugpy.breakpoint()

        # in case we missed logging closer to the source
        logger.exception(
            msg=exc_value.message, exc_info=(exc_type, exc_value, exc_traceback)
        )

        if telemetry.telemetry_data is not None:
            with contextlib.suppress(Exception):
                telemetry.telemetry_data.ato_error += 1
    elif issubclass(exc_type, BaseExceptionGroup):
        for e in exc_value.exceptions:
            _handle_exception(type(e), e, e.__traceback__)
    else:
        with contextlib.suppress(Exception):
            telemetry.telemetry_data.crash += 1
        logger.exception(
            "Uncaught compiler exception", exc_info=(exc_type, exc_value, exc_traceback)
        )


def handle_exception(exc_type, exc_value, exc_traceback):
    from atopile import telemetry

    try:
        _handle_exception(exc_type, exc_value, exc_traceback)
    except Exception as e:
        sys.__excepthook__(type(e), e, e.__traceback__)
    finally:
        with contextlib.suppress(Exception):
            telemetry.log_telemetry()

        rich.print(
            "\n\nUnfortunately errors ^^^ stopped the build. "
            "If you need a hand jump on [#9656ce]Discord[/]! [link=https://discord.gg/mjtxARsr9V]https://discord.gg/mjtxARsr9V[/] :wave:"  # noqa: E501  # pre-existing
        )
        sys.exit(1)


if not FLOG_FMT:
    sys.excepthook = handle_exception



from __future__ import annotations

import logging
from dataclasses import Field, dataclass, fields, is_dataclass
from enum import Enum, IntEnum, StrEnum
from os import PathLike
from pathlib import Path
from types import UnionType
from typing import TYPE_CHECKING, Any, Callable, Iterator, Union, get_args, get_origin

import sexpdata
from dataclasses_json import CatchAll
from dataclasses_json.utils import CatchAllVar
from sexpdata import Symbol

from faebryk.libs.exceptions import UserResourceException
from faebryk.libs.sexp.util import prettify_sexp_string
from faebryk.libs.util import (
    ConfigFlag,
    cast_assert,
    duplicates,
    groupby,
    zip_non_locked,
)

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

SEXP_LOG = ConfigFlag(
    "SEXP_LOG", default=False, descr="Enable sexp decode logging (very verbose)"
)


logger = logging.getLogger(__name__)

# TODO: Should be its own repo

"""
This is a generic sexp-dataclass converter (similar to dataclass_json).
It is used to convert between dataclasses and sexp (s-expressions).
It only supports a specific subset of sexp that is used by KiCAD with following rules:
- Only atom order is important
- Multi-key dict is supported
"""


@dataclass
class sexp_field(dict[str, Any]):
    """
    Metadata for a field to be used in the sexp conversion

    :param bool positional: If True, the fields position will be used instead
    of its name
    :param bool multidict: If True, the field will be converted to a list of
    key-value pairs. Not compatible in combination with positional.
    :param Callable key: Function to extract key from value in multidict
    :param Any assert_value: Assert that the value is equal to this value
    :param int order: Order of the field in the sexp, lower is first,
    can be less than 0. Only used if not positional.
    :param Callable[[Any], Any] | None preprocessor: Run before conversion
    """

    positional: bool = False
    multidict: bool = False
    key: Callable[[Any], Any] | None = None
    assert_value: Any | None = None
    order: int = 0
    preprocessor: Callable[[Any], Any] | None = None

    def __post_init__(self):
        super().__init__({"metadata": {"sexp": self}})

        assert not (self.positional and self.multidict)
        assert (self.key is None) or self.multidict, "Key only supported for multidict"

    @classmethod
    def from_field(cls, f: Field):
        out = f.metadata.get("sexp", cls())
        assert isinstance(out, cls)
        return out


class SymEnum(StrEnum): ...


class DecodeError(UserResourceException):
    """Error during decoding"""


def _prettify_stack(stack: list[tuple[str, type]] | None) -> str:
    if stack is None:
        return "<top-level>"
    return ".".join(s[0] for s in stack)


def _convert(
    val,
    t,
    stack: list[tuple[str, type]] | None = None,
    name: str | None = None,
    sp: sexp_field | None = None,
    ignore_assertions: bool = False,
):
    if name is None:
        name = "<" + t.__name__ + ">"
    if stack is None:
        stack = []
    substack = stack + [(name, t)]

    try:
        # Run preprocessor, if it exists
        if sp and sp.preprocessor:
            val = sp.preprocessor(val)

        # Recurse (GenericAlias e.g list[])
        if (origin := get_origin(t)) is not None:
            args = get_args(t)
            if origin is list:
                return [
                    _convert(
                        _val, args[0], substack, ignore_assertions=ignore_assertions
                    )
                    for _val in val
                ]
            if origin is tuple:
                return tuple(
                    _convert(_val, _t, substack, ignore_assertions=ignore_assertions)
                    for _val, _t in zip(val, args)
                )
            if (
                origin in (Union, UnionType)
                and len(args) == 2
                and args[1] is type(None)
            ):
                # Optional[T] == Union[T, None]
                return (
                    _convert(
                        val, args[0], substack, ignore_assertions=ignore_assertions
                    )
                    if val is not None
                    else None
                )

            raise NotImplementedError(f"{origin} not supported")

        #
        if is_dataclass(t):
            return _decode(val, t, substack, ignore_assertions=ignore_assertions)

        # Primitive

        # Unpack list if single atom
        if isinstance(val, list) and len(val) == 1 and not isinstance(val[0], list):
            val = val[0]

        if issubclass(t, bool):
            # See parseMaybeAbsentBool in kicad
            # Default: (hide) hide None
            # True: (hide yes)
            # False: (hide no)

            # hide, None -> automatically filtered

            # (hide yes) (hide no)
            if val in [Symbol("yes"), Symbol("no")]:
                return val == Symbol("yes")

            # (hide)
            if val == []:
                return None

            raise ValueError(f"Invalid value for bool: {val}")

        if isinstance(val, Symbol):
            return t(str(val))

        return t(val)
    except DecodeError:
        raise
    except Exception as e:
        raise DecodeError(
            f"Failed to decode {_prettify_stack(substack)} ({t}) with {val} "
        ) from e


netlist_obj = str | Symbol | int | float | bool | list
netlist_type = list[netlist_obj]


def _decode[T: DataclassInstance](
    sexp: netlist_type,
    t: type[T],
    stack: list[tuple[str, type]] | None = None,
    ignore_assertions: bool = False,
) -> T:
    if SEXP_LOG:
        logger.debug(f"parse into: {t.__name__} {'-'*40}")
        logger.debug(f"sexp: {sexp}")

    # check if t is dataclass type
    if not hasattr(t, "__dataclass_fields__"):
        # is_dataclass(t) trips mypy
        raise TypeError(f"{t} is not a dataclass type")

    value_dict = {}
    unprocessed = {}

    # Fields
    fs = fields(t)
    key_fields = {f.name: f for f in fs if not sexp_field.from_field(f).positional}
    positional_fields = {
        i: f for i, f in enumerate(fs) if sexp_field.from_field(f).positional
    }
    catch_all_fields = [
        f for f in fields(t) if f.type is CatchAllVar or f.type is CatchAll
    ]
    if len(catch_all_fields) == 0:
        pass
    elif len(catch_all_fields) == 1:
        value_dict[catch_all_fields[0].name] = unprocessed
    elif len(catch_all_fields) > 1:
        raise ValueError(f"Multiple catch-all fields not allowed: {catch_all_fields}")

    # Values
    ungrouped_key_values = []
    # I'd prefer to do this through a filter/comprehension, but I don't see a good way
    for i, val in enumerate(sexp):
        if isinstance(val, list):
            if len(val):
                if isinstance(key := val[0], Symbol):
                    if str(key) in key_fields or str(key) + "s" in key_fields:
                        ungrouped_key_values.append(val)
                        continue

        unprocessed[i] = val

    key_values = groupby(
        ungrouped_key_values,
        lambda val: (
            str(val[0]) + "s" if str(val[0]) + "s" in key_fields else str(val[0])
        ),
    )
    pos_values = {
        i: val
        for i, val in enumerate(sexp)
        if isinstance(val, (str, int, float, Symbol, bool))
        or (isinstance(val, list) and (not len(val) or not isinstance(val[0], Symbol)))
        # and i in positional_fields
        # and positional_fields[i].name not in value_dict
    }

    if SEXP_LOG:
        logger.debug(f"processing: {_prettify_stack(stack)}")
        logger.debug(f"key_fields: {list(key_fields.keys())}")
        logger.debug(
            f"positional_fields: {list(f.name for f in positional_fields.values())}"
        )
        logger.debug(f"key_values: {list(key_values.keys())}")
        logger.debug(f"pos_values: {pos_values}")
        logger.debug(f"maybe unprocessed values (or positional): {unprocessed}")

    # Parse --------------------------------------------------------------

    # Key-Value
    for s_name, f in key_fields.items():
        name = f.name
        sp = sexp_field.from_field(f)
        if s_name not in key_values:
            if sp.multidict and not (f.default_factory or f.default):
                base_type = get_origin(f.type) or f.type
                value_dict[name] = base_type()
            # will be automatically filled by factory
            continue

        values = key_values[s_name]
        if sp.multidict:
            origin = get_origin(f.type)
            args = get_args(f.type)
            if origin is list:
                val_t = args[0]
                value_dict[name] = [
                    _convert(
                        _val[1:],
                        val_t,
                        stack,
                        name,
                        sp,
                        ignore_assertions=ignore_assertions,
                    )
                    for _val in values
                ]
            elif origin is dict:
                if not sp.key:
                    raise ValueError(f"Key function required for multidict: {f.name}")
                key_t = args[0]
                val_t = args[1]
                converted_values = [
                    _convert(
                        _val[1:],
                        val_t,
                        stack,
                        name,
                        sp,
                        ignore_assertions=ignore_assertions,
                    )
                    for _val in values
                ]
                values_with_key = [(sp.key(_val), _val) for _val in converted_values]

                if not all(isinstance(k, key_t) for k, _ in values_with_key):
                    raise KeyError(
                        f"Key function returned invalid type in field {f.name}:"
                        f" {key_t=} types={[v[0] for v in values_with_key]}"
                    )
                if d := duplicates(values_with_key, key=lambda v: v[0]):
                    raise ValueError(f"Duplicate keys: {d}")
                value_dict[name] = dict(values_with_key)
            else:
                raise NotImplementedError(
                    f"Multidict not supported for {origin} in field {f}"
                )
        else:
            assert len(values) == 1, f"Duplicate key: {name}"
            out = _convert(
                values[0][1:],
                f.type,
                stack,
                name,
                sp,
                ignore_assertions=ignore_assertions,
            )
            # if val is None, use default
            if out is not None:
                value_dict[name] = out

    # Positional
    for f, (sexp_i, v) in (
        it := zip_non_locked(positional_fields.values(), pos_values.items())
    ):
        sp = sexp_field.from_field(f)
        # special case for missing positional empty StrEnum fields
        if (
            isinstance(f.type, type)
            and issubclass(f.type, StrEnum)
            and "" in f.type
            and not isinstance(v, Symbol)
        ):
            out = _convert(
                Symbol(""),
                f.type,
                stack,
                f.name,
                sp,
                ignore_assertions=ignore_assertions,
            )
            # only advance field iterator
            # if no more positional fields, there shouldn't be any more values
            if it.next(0) is None:
                raise ValueError(f"Unexpected symbol {v}")
        else:
            # positional list = var args
            origin = get_origin(f.type)
            if origin is list:
                vs = [v]
                # consume all values
                while (next_val := it.next(1, None)) is not None:
                    vs.append(next_val[1])

                out = _convert(
                    vs, f.type, stack, f.name, sp, ignore_assertions=ignore_assertions
                )
            else:
                out = _convert(
                    v, f.type, stack, f.name, sp, ignore_assertions=ignore_assertions
                )

        unprocessed.pop(sexp_i, None)
        value_dict[f.name] = out

    # Check assertions ----------------------------------------------------
    for f in fs:
        sp = sexp_field.from_field(f)
        if sp.assert_value is not None:
            assert value_dict[f.name] == sp.assert_value, (
                f"Fileformat assertion! {f.name} has to be"
                f" {sp.assert_value} but is {value_dict[f.name]}"
            )

    # Unprocessed values should be None if there aren't any,
    # so they don't get reproduced etc...
    if catch_all_fields and not unprocessed:
        value_dict[catch_all_fields[0].name] = None

    if SEXP_LOG:
        logger.debug(f"value_dict: {value_dict}")

    try:
        out = t(**value_dict)
    except TypeError as e:
        raise TypeError(f"Failed to create {t} with {value_dict}") from e

    # set parent pointers for all dataclasses in the tree
    for v in value_dict.values():
        if isinstance(v, list):
            vs = v
        else:
            vs = [v]
        for v_ in vs:
            if is_dataclass(v_):
                setattr(v_, "_parent", out)

    return out


def _convert2(val: Any) -> netlist_obj | None:
    if val is None:
        return None
    if is_dataclass(val):
        return _encode(val)
    if isinstance(val, list):
        return [_convert2(v) for v in val]
    if isinstance(val, tuple):
        return [_convert2(v) for v in val]
    if isinstance(val, dict):
        return [_convert2(v) for v in val.values()]
    if isinstance(val, SymEnum):
        return Symbol(val)
    if isinstance(val, StrEnum):
        return str(val)
    if isinstance(val, IntEnum):
        return int(val)
    if isinstance(val, Enum):
        return Symbol(val)
    if isinstance(val, bool):
        return Symbol("yes" if val else "no")
    if isinstance(val, float):
        if val.is_integer():
            return int(val)
        return val
    if isinstance(val, (str, int)):
        return val

    return str(val)


def _encode(t) -> netlist_type:
    if not is_dataclass(t):
        raise TypeError(f"{t} is not a dataclass type")

    sexp: netlist_type = []

    def _append(_val):
        if val is None:
            return
        sexp.append(_val)

    catch_all_fields = [
        f for f in fields(t) if f.type is CatchAllVar or f.type is CatchAll
    ]
    if len(catch_all_fields) == 0:
        catch_all_field = None
    elif len(catch_all_fields) == 1:
        catch_all_field = catch_all_fields[0]
    else:
        raise ValueError(f"Multiple catch-all fields not allowed: {catch_all_fields}")

    fs = [(f, sexp_field.from_field(f)) for f in fields(t) if f != catch_all_field]

    for f, sp in sorted(fs, key=lambda x: (not x[1].positional, x[1].order)):
        name = f.name
        val = getattr(t, name)

        if sp.positional:
            if isinstance(val, list):
                for v in val:
                    _append(_convert2(v))
                continue
            _append(_convert2(val))
            continue

        def _append_kv(name, v):
            converted = _convert2(v)
            if converted is None:
                return
            if isinstance(converted, list):
                _append([Symbol(name), *converted])
                return
            _append([Symbol(name), converted])

        if sp.multidict:
            if isinstance(val, list):
                assert get_origin(f.type) is list
                _val = val
            elif isinstance(val, dict):
                assert get_origin(f.type) is dict
                _val = val
                _val = val.values()
            else:
                raise TypeError()
            for v in _val:
                _append_kv(f.name.removesuffix("s"), v)
        else:
            _append_kv(f.name, val)

    if catch_all_field is not None:
        catch_all: dict[int, Any] = getattr(t, catch_all_field.name, {}) or {}
        for i, v in sorted(catch_all.items(), key=lambda x: x[0]):
            sexp.insert(i, v)

    if SEXP_LOG:
        logger.debug(f"Dumping {type(t).__name__} {'-'*40}")
        logger.debug(f"Obj: {t}")
        logger.debug(f"Sexp: {sexp}")

    return sexp


def loads[T: DataclassInstance](
    s: str | Path | list, t: type[T], ignore_assertions: bool = False
) -> T:
    text = s
    sexp = s
    if isinstance(s, Path):
        text = s.read_text()
    if isinstance(text, str):
        try:
            sexp = sexpdata.loads(text)
        except Exception as e:
            raise DecodeError(f"Failed to parse sexp: {text}") from e

    return _decode([sexp], t, ignore_assertions=ignore_assertions)


def dumps(obj, path: PathLike | None = None) -> str:
    path = Path(path) if path else None
    sexp = _encode(obj)[0]
    text = sexpdata.dumps(sexp)
    text = prettify_sexp_string(text)
    if path:
        path.write_text(text)
    return text


def dump_single(obj) -> str:
    @dataclass
    class _(SEXP_File):
        node: type[obj]

    filenode = _(node=obj)

    return filenode.dumps()


class SEXP_File:
    @classmethod
    def loads(cls, path_or_string_or_data: Path | str | list):
        return loads(path_or_string_or_data, cls)

    def dumps(self, path: PathLike | None = None):
        return dumps(self, path)


def get_parent[T](obj, t: type[T]) -> T:
    assert hasattr(obj, "_parent")
    return cast_assert(t, obj._parent)


# TODO move
class JSON_File:
    @classmethod
    def loads[T](cls: type[T], path: Path | str) -> T:
        text = path
        if isinstance(path, Path):
            text = path.read_text()
        return cls.from_json(text)

    def dumps(self, path: PathLike | None = None):
        path = Path(path) if path else None
        text = self.to_json(indent=4)
        if path:
            path.write_text(text)
        return text


def dataclass_dfs(obj) -> Iterator[tuple[Any, list, list[str]]]:
    return _iterate_tree(obj, [], [])


def _iterate_tree(
    obj, path: list, name_path: list[str]
) -> Iterator[tuple[Any, list, list[str]]]:
    out_path = path + [obj]

    yield obj, path, name_path

    if is_dataclass(obj):
        for f in fields(obj):
            yield from _iterate_tree(
                getattr(obj, f.name), out_path, name_path + [f".{f.name}"]
            )
    elif isinstance(obj, list):
        for i, v in enumerate(obj):
            yield from _iterate_tree(v, out_path, name_path + [f"[{i}]"])
    elif isinstance(obj, dict):
        for k, v in obj.items():
            yield from _iterate_tree(v, out_path, name_path + [f"[{repr(k)}]"])



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
import logging
from enum import Enum

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class I2C(ModuleInterface):
    scl: F.ElectricLogic
    sda: F.ElectricLogic

    frequency = L.p_field(
        units=P.Hz,
        likely_constrained=True,
        soft_set=L.Range(10 * P.kHz, 3.4 * P.MHz),
    )

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    def terminate(self, owner: Module):
        # TODO: https://www.ti.com/lit/an/slva689/slva689.pdf

        self.pull_up_sda = self.sda.pulled.pull(up=True, owner=owner)
        self.pull_up_scl = self.scl.pulled.pull(up=True, owner=owner)

    class SpeedMode(Enum):
        low_speed = 10 * P.khertz
        standard_speed = 100 * P.khertz
        fast_speed = 400 * P.khertz
        high_speed = 3.4 * P.Mhertz

    @staticmethod
    def define_max_frequency_capability(mode: SpeedMode):
        return L.Range(I2C.SpeedMode.low_speed.value, mode.value)

    def __preinit__(self) -> None:
        self.frequency.add(F.is_bus_parameter())



# excepthook must be installed before typer is imported
import atopile.cli.excepthook  # noqa: F401, I001

import json
import logging
import sys
from enum import Enum
from importlib.metadata import version as get_package_version
from pathlib import Path
from typing import Annotated

import typer

from atopile import telemetry, version
from atopile.cli import build, configure, create, inspect, install, view
from atopile.cli.logging import handler, logger
from atopile.config import config
from atopile.version import check_for_update
from faebryk.libs.logging import FLOG_FMT

app = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=bool(FLOG_FMT),  # required to override the excepthook
    rich_markup_mode="rich",
)


def python_interpreter_path(ctx: typer.Context, value: bool):
    """Print the current python interpreter path."""
    if not value or ctx.resilient_parsing:
        return
    typer.echo(sys.executable)
    raise typer.Exit()


def atopile_src_path(ctx: typer.Context, value: bool):
    """Print the current python interpreter path."""
    if not value or ctx.resilient_parsing:
        return
    typer.echo(Path(__file__).parent.parent)
    raise typer.Exit()


def version_callback(ctx: typer.Context, value: bool):
    """Output a version string meeting the pypa version spec."""
    if not value or ctx.resilient_parsing:
        return
    typer.echo(get_package_version("atopile"))
    raise typer.Exit()


def semver_callback(ctx: typer.Context, value: bool):
    """Output a version string meeting the semver.org spec."""
    if not value or ctx.resilient_parsing:
        return
    version_string = get_package_version("atopile")
    typer.echo(version.parse(version_string))
    raise typer.Exit()


@app.callback()
def cli(
    ctx: typer.Context,
    non_interactive: Annotated[
        bool | None,
        typer.Option(
            "--non-interactive", envvar=["ATO_NON_INTERACTIVE", "NONINTERACTIVE"]
        ),
    ] = None,
    debug: Annotated[
        bool,
        typer.Option("--debug", help="Wait to attach debugger on start"),
    ] = False,
    verbose: Annotated[
        int,
        typer.Option("--verbose", "-v", count=True, help="Increase verbosity"),
    ] = 0,
    python_path: Annotated[
        bool, typer.Option(hidden=True, callback=python_interpreter_path)
    ] = False,
    atopile_path: Annotated[
        bool, typer.Option(hidden=True, callback=atopile_src_path)
    ] = False,
    version: Annotated[
        bool | None,
        typer.Option("--version", callback=version_callback, is_eager=True),
    ] = None,
    semver: Annotated[
        bool | None,
        typer.Option("--semver", callback=semver_callback, is_eager=True),
    ] = None,
):
    if debug:
        import debugpy  # pylint: disable=import-outside-toplevel

        debug_port = 5678
        debugpy.listen(("localhost", debug_port))
        logger.info("Starting debugpy on port %s", debug_port)
        debugpy.wait_for_client()

    # set the log level
    if verbose == 1:
        handler.hide_traceback_types = ()
        handler.tracebacks_show_locals = True
    elif verbose == 2:
        handler.tracebacks_suppress_map = {}  # Traceback through atopile infra
    elif verbose >= 3:
        logger.root.setLevel(logging.DEBUG)
        handler.traceback_level = logging.WARNING

    # FIXME: this won't work properly when configs
    # are reloaded from a pointed-to file (eg in `ato build path/to/file`)
    # from outside a project directory
    if non_interactive is not None:
        config.interactive = not non_interactive

    if ctx.invoked_subcommand:
        check_for_update()

        # Initialize telemetry
        telemetry.setup_telemetry_data(ctx.invoked_subcommand)

    if config.interactive and ctx.invoked_subcommand != "configure":
        configure.do_configure_if_needed()


app.command()(build.build)
app.add_typer(create.create_app, name="create")
app.command()(install.install)
app.command()(configure.configure)
app.command()(inspect.inspect)
app.command()(view.view)


@app.command(hidden=True)
def export_config_schema(pretty: bool = False):
    from atopile.config import ProjectConfig

    config_schema = ProjectConfig.model_json_schema()

    if pretty:
        print(json.dumps(config_schema, indent=4))
    else:
        print(json.dumps(config_schema))


class ConfigFormat(str, Enum):
    python = "python"
    json = "json"


@app.command(hidden=True)
def dump_config(format: ConfigFormat = ConfigFormat.python):
    from rich import print

    print(config.project.model_dump(mode=format))


def main():
    app()


if __name__ == "__main__":
    main()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P


class USB_C_5V_PSU(Module):
    # interfaces
    power_out: F.ElectricPower
    usb: F.USB_C

    # components
    configuration_resistors = L.list_field(
        2,
        lambda: F.Resistor().builder(
            lambda r: r.resistance.constrain_subset(
                L.Range.from_center_rel(5.1 * P.kohm, 0.05)
            )
        ),
    )

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )

    def __preinit__(self):
        # configure as ufp with 5V@max3A
        self.usb.cc1.connect_via(self.configuration_resistors[0], self.power_out.lv)
        self.usb.cc2.connect_via(self.configuration_resistors[1], self.power_out.lv)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class can_attach_to_footprint_symmetrically(F.can_attach_to_footprint.impl()):
    def attach(self, footprint: F.Footprint):
        self.obj.add(F.has_footprint_defined(footprint))

        for i, j in zip(
            footprint.get_children(direct_only=True, types=F.Pad),
            self.obj.get_children(direct_only=True, types=F.Electrical),
        ):
            i.attach(j)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from enum import Enum, auto

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class OLED_Module(Module):
    class DisplayResolution(Enum):
        H64xV32 = auto()
        H128xV32 = auto()
        H128xV64 = auto()
        H256xV64 = auto()

    class DisplaySize(Enum):
        INCH_0_96 = auto()
        INCH_1_12 = auto()
        INCH_1_27 = auto()
        INCH_1_3 = auto()
        INCH_1_5 = auto()
        INCH_2_23 = auto()
        INCH_2_3 = auto()
        INCH_2_42 = auto()
        INCH_2_7 = auto()

    class DisplayController(Enum):
        SSD1315 = auto()
        SSD1306 = auto()
        SSD1309 = auto()

    power: F.ElectricPower
    i2c: F.I2C

    display_resolution = L.p_field(domain=L.Domains.ENUM(DisplayResolution))
    display_controller = L.p_field(domain=L.Domains.ENUM(DisplayController))
    display_size = L.p_field(domain=L.Domains.ENUM(DisplaySize))

    def __preinit__(self):
        self.power.voltage.constrain_subset(L.Range(3.0 * P.V, 5 * P.V))

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.DS
    )

    @L.rt_field
    def can_be_decoupled(self):
        class _(F.can_be_decoupled.impl()):
            def decouple(self, owner: Module):
                obj = self.get_obj(OLED_Module)
                obj.power.decoupled.decouple(owner=owner).capacitance.constrain_subset(
                    L.Range(100 * P.uF, 220 * P.uF)
                )

        return _()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class Common_Mode_Filter(Module):
    coil_a: F.Inductor
    coil_b: F.Inductor

    inductance = L.p_field(
        units=P.H,
        likely_constrained=True,
        soft_set=L.Range(1 * P.µH, 10 * P.mH),
        tolerance_guess=10 * P.percent,
    )
    self_resonant_frequency = L.p_field(
        units=P.Hz,
        likely_constrained=True,
        soft_set=L.Range(100 * P.Hz, 1 * P.MHz),
        tolerance_guess=10 * P.percent,
    )
    max_current = L.p_field(
        units=P.A,
        likely_constrained=True,
        soft_set=L.Range(1 * P.A, 10 * P.A),
    )
    dc_resistance = L.p_field(
        units=P.Ω,
    )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.FL
    )

    def __preinit__(self):
        # ----------------------------------------
        #            parametrization
        # ----------------------------------------
        for coil in [self.coil_a, self.coil_b]:
            coil.inductance.alias_is(self.inductance)
            coil.self_resonant_frequency.alias_is(self.self_resonant_frequency)
            coil.max_current.alias_is(self.max_current)
            coil.dc_resistance.alias_is(self.dc_resistance)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from pathlib import Path

import faebryk

logger = logging.getLogger(__name__)

_RESOURCES_SUFFIX = "test/common/resources"


def ROOT_PATH() -> Path:
    editable_root = Path(faebryk.__file__).parent.parent.parent
    if (editable_root / _RESOURCES_SUFFIX).exists():
        return editable_root
    cwd_root = Path.cwd()
    if (cwd_root / _RESOURCES_SUFFIX).exists():
        return cwd_root
    raise FileNotFoundError("Could not find root directory")


RESOURCES_PATH = ROOT_PATH() / _RESOURCES_SUFFIX
FILEFORMATS_PATH = RESOURCES_PATH / "fileformats/kicad"


def _VERSION_DIR(version: int) -> Path:
    return FILEFORMATS_PATH / f"v{version}"


DEFAULT_VERSION = 8


def _FP_DIR(version: int = DEFAULT_VERSION) -> Path:
    return _VERSION_DIR(version) / "fp"


def _NETLIST_DIR(version: int = DEFAULT_VERSION) -> Path:
    return _VERSION_DIR(version) / "netlist"


def _SCH_DIR(version: int = DEFAULT_VERSION) -> Path:
    return _VERSION_DIR(version) / "sch"


def _SYM_DIR(version: int = DEFAULT_VERSION) -> Path:
    return _VERSION_DIR(version) / "sym"


def _PRJ_DIR(version: int = DEFAULT_VERSION) -> Path:
    return _VERSION_DIR(version) / "prj"


def _FPLIB_DIR(version: int = DEFAULT_VERSION) -> Path:
    return _VERSION_DIR(version) / "fplib"


def _PCB_DIR(version: int = DEFAULT_VERSION) -> Path:
    return _VERSION_DIR(version) / "pcb"


PRJFILE = _PRJ_DIR() / "test.kicad_pro"
PCBFILE = _PCB_DIR() / "test.kicad_pcb"
FPFILE = _FP_DIR() / "test.kicad_mod"
NETFILE = _NETLIST_DIR() / "test_e.net"
FPLIBFILE = _FPLIB_DIR(7) / "fp-lib-table"
SCHFILE = _SCH_DIR() / "test.kicad_sch"
SYMFILE = _SYM_DIR() / "test.kicad_sym"



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from typing import Callable

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L


class GenericBusProtection(Module):
    type T = ModuleInterface

    @L.rt_field
    def bus_unprotected(self):
        return self.bus_factory()

    @L.rt_field
    def bus_protected(self):
        return self.bus_factory()

    def __init__(self, bus_factory: Callable[[], T]) -> None:
        super().__init__()
        self.bus_factory = bus_factory

    def __preinit__(self):
        def get_mifs[U: ModuleInterface](
            bus: "GenericBusProtection.T", mif_type: type[U]
        ) -> set[U]:
            return bus.get_children(direct_only=True, types=mif_type)

        raw = list(
            zip(
                get_mifs(self.bus_unprotected, F.Electrical),
                get_mifs(self.bus_protected, F.Electrical),
            )
        )
        signals = list(
            zip(
                get_mifs(self.bus_unprotected, F.ElectricLogic),
                get_mifs(self.bus_protected, F.ElectricLogic),
            )
        )
        power = list(
            zip(
                get_mifs(self.bus_unprotected, F.ElectricPower),
                get_mifs(self.bus_protected, F.ElectricPower),
            )
        )

        fuse = L.list_field(len(power), F.Fuse)

        # Pass through except hv
        for power_unprotected, power_protected in power:
            power_unprotected.lv.connect(power_protected.lv)
        for logic_unprotected, logic_protected in signals:
            logic_unprotected.connect_shallow(logic_protected, signal=True, lv=True)
        for raw_unprotected, raw_protected in raw:
            raw_unprotected.connect(raw_protected)

        # Fuse
        for (power_unprotected, power_protected), fuse in zip(power, fuse):
            power_unprotected.hv.connect_via(fuse, power_protected.hv)
            # TODO maybe shallow connect?
            power_protected.voltage.alias_is(power_unprotected.voltage)

        # TVS
        if self.bus_protected.has_trait(F.can_be_surge_protected):
            self.bus_protected.get_trait(F.can_be_surge_protected).protect(self)
        else:
            for line_unprotected, line_protected in signals + power + raw:
                line_protected.get_trait(F.can_be_surge_protected).protect(self)

        # TODO add shallow connect

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.bus_unprotected, self.bus_protected)



from atopile.datatypes import KeyOptItem, KeyOptMap, Ref, StackList


def test_ref_from_one():
    assert Ref.from_one("foo") == ("foo",)
    assert Ref.from_one(42) == ("42",)


def test_ref_add_name():
    assert Ref.from_one("foo").add_name("bar") == ("foo", "bar")


def test_keyoptitem_from_kv():
    assert KeyOptItem.from_kv(None, "foo") == (None, "foo")


def test_keyoptmap_from_item():
    assert KeyOptMap.from_item(KeyOptItem.from_kv(None, "foo")) == ((None, "foo"),)


def test_keyoptmap_from_kv():
    assert KeyOptMap.from_kv(None, "foo") == ((None, "foo"),)


def test_keyoptitem_ref():
    assert KeyOptItem((None, "foo")).ref is None
    assert KeyOptItem((Ref.from_one("foo"), "bar")).ref == ("foo",)


def test_stack_list():
    stack = StackList()
    with stack.enter(1):
        assert stack == [1]
        assert stack.top == 1
        with stack.enter(2):
            assert stack == [1, 2]
            assert stack.top == 2
        assert stack == [1]
        assert stack.top == 1



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import math
from enum import Enum, auto

import pytest

from faebryk.libs.library.L import DiscreteSet, EmptySet, Range, RangeWithGaps, Single
from faebryk.libs.sets.numeric_sets import (
    Numeric_Interval,
    Numeric_Interval_Disjoint,
    float_round,
    rel_round,
)
from faebryk.libs.sets.quantity_sets import (
    Quantity_Interval,
    Quantity_Interval_Disjoint,
    Quantity_Set_Discrete,
)
from faebryk.libs.sets.sets import BoolSet, EnumSet, P_Set
from faebryk.libs.units import P, Unit, dimensionless, quantity
from faebryk.libs.util import cast_assert


def test_interval_intersection_simple():
    x = Range(0, 10)
    y = x & Range(5, 15)
    assert y == Range(5, 10)


def test_interval_intersection_empty():
    x = Range(0, 10)
    y = x & Range(15, 20)
    assert y == EmptySet(dimensionless)


def test_interval_unit_none():
    x = Range(0, 10)
    assert not x.units.is_compatible_with(P.V)


def test_interval_unit_same():
    y = Range(0 * P.V, 10 * P.V)
    assert y.units.is_compatible_with(P.V)


def test_interval_unit_different():
    with pytest.raises(ValueError):
        Range(0 * P.V, 10 * P.A)
    with pytest.raises(ValueError):
        Range(0 * P.V, 10 * P.V, units=cast_assert(Unit, P.A))
    with pytest.raises(ValueError):
        Range(max=10 * P.V, units=cast_assert(Unit, P.A))
    with pytest.raises(ValueError):
        Range(min=10 * P.V, units=cast_assert(Unit, P.A))


def test_set_min_elem():
    x = DiscreteSet(5, 3, 2, 4, 1)
    assert x.min_elem == 1


def test_set_closest_elem():
    x = RangeWithGaps((5, 6), (7, 8), DiscreteSet(2, 4, 1))
    assert x.closest_elem(quantity(0)) == 1
    assert x.closest_elem(quantity(1)) == 1
    assert x.closest_elem(quantity(5.1)) == 5.1
    assert x.closest_elem(quantity(4.9)) == 5
    assert x.closest_elem(quantity(4.1)) == 4
    assert x.closest_elem(quantity(6.9)) == 7


def test_set_contains():
    x = DiscreteSet(5, 3, 2, 4, 1)
    assert 3 * dimensionless in x
    assert 6 * dimensionless not in x


def test_union_min_elem():
    x = RangeWithGaps(
        (4, 5),
        (3, 7),
        Single(9),
        RangeWithGaps(Range(1, 2), RangeWithGaps(Range(0, 1))),
    )
    assert x.min_elem == 0


def test_union_contains():
    x = RangeWithGaps(
        (4, 5),
        (3, 7),
        Single(9),
        RangeWithGaps((1, 2), RangeWithGaps((0, 1))),
    )
    assert 0 * dimensionless in x
    assert 1 * dimensionless in x
    assert 2 * dimensionless in x
    assert 3 * dimensionless in x
    assert 4 * dimensionless in x
    assert 5 * dimensionless in x
    assert 6 * dimensionless in x
    assert 7 * dimensionless in x
    assert 8 * dimensionless not in x
    assert 9 * dimensionless in x
    assert 10 * dimensionless not in x

    x = RangeWithGaps(Range(max=1.5 * P.V), Range(2.5 * P.V, 3.5 * P.V))
    assert float("-inf") * P.V in x
    assert 1 * P.V in x
    assert 1.5 * P.V in x
    assert 2 * P.V not in x
    assert 2.5 * P.V in x
    assert 3 * P.V in x
    assert 3.5 * P.V in x
    assert 4 * P.V not in x
    assert float("inf") * P.V not in x
    assert 1 not in x
    assert 1 * dimensionless not in x


def test_union_empty():
    x = RangeWithGaps(
        EmptySet(dimensionless),
        RangeWithGaps(EmptySet(dimensionless), DiscreteSet(units=dimensionless)),
    )
    assert x.is_empty()


def test_add_empty():
    assert EmptySet(dimensionless) + RangeWithGaps((0, 1)) == EmptySet(dimensionless)


def test_addition():
    assert Range(0, 1) + Range(2, 3) == Range(2, 4)
    assert Range(0, 1) + Single(2) == Range(2, 3)
    assert RangeWithGaps(Single(2), Single(3)) + RangeWithGaps((0, 1)) == Range(2, 4)
    assert RangeWithGaps(Single(10), Range(20, 21)) + RangeWithGaps(
        (0, 1), (100, 101)
    ) == RangeWithGaps((10, 11), (110, 111), (20, 22), (120, 122))


def test_addition_unit():
    assert Range(0 * P.V, 1 * P.V) + Range(2 * P.V, 3 * P.V) == Range(2 * P.V, 4 * P.V)


def test_subtraction():
    assert Range(0, 1) - Range(2, 3) == Range(-3, -1)
    assert Range(0, 1) - Single(2) == Range(-2, -1)


def test_subtraction_unit():
    assert Range(0 * P.V, 1 * P.V) - Range(2 * P.V, 3 * P.V) == Range(
        -3 * P.V, -1 * P.V
    )


def test_multiplication():
    assert Range(0, 2) * Range(2, 3) == Range(0, 6)
    assert Range(0, 1) * Single(2) == Range(0, 2)
    assert Range(0, 1) * Single(-2) == Range(-2, 0)
    assert Range(-1, 1) * Range(2, 4) == Range(-4, 4)
    assert DiscreteSet(0, 1) * DiscreteSet(2, 3) == DiscreteSet(0, 2, 3)
    assert DiscreteSet(0, 1) * DiscreteSet(2, 3) * RangeWithGaps(
        (-1, 0)
    ) == RangeWithGaps((0, 0), (-2, 0), (-3, 0))


def test_multiplication_unit():
    assert Range(0 * P.V, 2 * P.V) * Range(2 * P.A, 3 * P.A) == Range(0 * P.W, 6 * P.W)


def test_invert():
    assert Range(1, 2).op_invert() == Range(0.5, 1)
    assert Range(-2, -1).op_invert() == Range(-1, -0.5)
    assert Range(-1, 1).op_invert() == RangeWithGaps(
        (float("-inf"), -1), (1, float("inf"))
    )
    assert RangeWithGaps((-4, 2), (-1, 3)).op_invert() == RangeWithGaps(
        Range(max=-0.25), Range(min=1 / 3)
    )


def test_invert_unit():
    assert Range(1 * P.V, 2 * P.V).op_invert() == Range(1 / (2 * P.V), 1 / (1 * P.V))


def test_division():
    assert Range(0, 1) / Range(2, 3) == Range(0, 0.5)
    assert Range(0, 1) / Range(0, 3) == Range(min=0.0)


def test_division_unit():
    assert Range(0 * P.V, 1 * P.V) / Range(2 * P.A, 3 * P.A) == Range(
        0 * P.ohm, 1 / 2 * P.ohm
    )


def test_pow():
    assert RangeWithGaps((0, 1), (2, 3)) ** Range(2, 3) == RangeWithGaps(
        (0, 1), (4, 27)
    )


@pytest.mark.parametrize(
    "base,exp, expected",
    # [-11, 10]^2 -> [0, 11^2]
    # [-11, 10]^3 -> [-11^3, 10^3]
    # [-11, 10]^[2,3] -> [-11^3, 10^3]
    # [-5, 2]^[2,3] -> [-5^3, -5^2]
    [
        (
            Range(-11, 10),
            Range(2, 2),
            Range(0, 11**2),
        ),
        (
            Range(-11, 10),
            Range(3, 3),
            Range(-(11**3), 10**3),
        ),
        (
            Range(-11, 10),
            DiscreteSet(2, 3),
            Range(-(11**3), 10**3),
        ),
        (
            Range(-5, 2),
            DiscreteSet(2, 3),
            Range(-(5**3), (5**2)),
        ),
    ],
)
def test_pow_extended(
    base: Range,
    exp: Range,
    expected: Range,
):
    base_q, exp_q, expected_q = (
        RangeWithGaps.from_value(base),
        RangeWithGaps.from_value(exp),
        RangeWithGaps.from_value(expected),
    )
    assert base_q**exp_q == expected_q


@pytest.mark.skip(
    "Zero crossing not implemented https://github.com/atopile/atopile/issues/614"
)
def test_pow_unit():
    assert RangeWithGaps((-3 * P.m, 1 * P.m)) ** Single(quantity(2)) == RangeWithGaps(
        (1 * P.m**2, 9 * P.m**2)
    )


@pytest.mark.skip(
    "Zero crossing not implemented https://github.com/atopile/atopile/issues/614"
)
def test_pow_div_eq():
    x = RangeWithGaps(Range(-5, 10))
    y = RangeWithGaps(Range(-2, 3))
    assert x / y == x * y**-1


def test_boolset_contain():
    assert BoolSet(True) in [BoolSet(True), BoolSet(False), BoolSet(True, False)]
    assert BoolSet(False) in [BoolSet(True), BoolSet(False), BoolSet(True, False)]
    assert BoolSet(True) not in [BoolSet(False), BoolSet(True, False)]
    assert BoolSet(False) not in [BoolSet(True), BoolSet(True, False)]
    assert True in [BoolSet(True)]
    assert False in [BoolSet(False)]
    assert True not in [BoolSet(True, False)]
    assert False not in [BoolSet(True, False)]
    assert True in BoolSet(True, False)
    assert False in BoolSet(True, False)


def test_boolset_eq():
    assert BoolSet(True) == BoolSet(True)
    assert BoolSet(False) == BoolSet(False)
    assert BoolSet(True, False) == BoolSet(True, False)
    assert BoolSet(True) != BoolSet(False)
    assert BoolSet(True) != BoolSet(True, False)
    assert True == BoolSet(True)  # noqa: E712
    with pytest.raises(Exception):
        bool(BoolSet(True))
    assert False == BoolSet(False)  # noqa: E712
    assert True != BoolSet(False)  # noqa: E712
    assert False != BoolSet(True)  # noqa: E712
    assert BoolSet(True, False) == BoolSet(False, True)
    assert True is not BoolSet(True)


def test_comparison():
    assert (
        RangeWithGaps((1 * P.V, 2 * P.V)) >= RangeWithGaps((0 * P.V, 1 * P.V))
    ) == BoolSet(True)
    assert (
        RangeWithGaps((1 * P.V, 2 * P.V)) <= RangeWithGaps((2 * P.V, 3 * P.V))
    ) == BoolSet(True)
    assert (
        RangeWithGaps((1 * P.V, 2 * P.V)) >= RangeWithGaps((0 * P.V, 0.5 * P.V))
    ) == BoolSet(True)
    assert (
        RangeWithGaps((1 * P.V, 2 * P.V)) <= RangeWithGaps((2.5 * P.V, 3 * P.V))
    ) == BoolSet(True)

    assert (
        RangeWithGaps((1 * P.V, 2 * P.V)) <= RangeWithGaps((0 * P.V, 0.5 * P.V))
    ) == BoolSet(False)
    assert (
        RangeWithGaps((1 * P.V, 2 * P.V)) >= RangeWithGaps((2.5 * P.V, 3 * P.V))
    ) == BoolSet(False)

    assert (
        RangeWithGaps((1 * P.V, 2 * P.V)) <= RangeWithGaps((0 * P.V, 1 * P.V))
    ) == BoolSet(True, False)
    assert (
        RangeWithGaps((1 * P.V, 2 * P.V)) >= RangeWithGaps((2 * P.V, 3 * P.V))
    ) == BoolSet(True, False)


def test_enumset():
    class E(Enum):
        A = auto()
        B = auto()
        C = auto()

    x = EnumSet(E.B, E.C)
    y = EnumSet(E.A, E.B)
    z = x & y

    assert z.enum.enum is E

    assert E.B in z
    assert E.A not in z

    assert z == EnumSet(E.B)


class _Enum(Enum):
    A = auto()
    B = auto()
    C = auto()


@pytest.mark.parametrize(
    "input_set,expected",
    [
        (Numeric_Interval(-10, 20), None),
        (Numeric_Interval(-10.5, 20.5), None),
        (Numeric_Interval(-10.5, 20), None),
        (Numeric_Interval(-math.inf, 50), None),
        (
            Numeric_Interval_Disjoint(
                Numeric_Interval(-10, 20), Numeric_Interval(30, 40.0)
            ),
            None,
        ),
        (BoolSet(True), None),
        (BoolSet.unbounded(), None),
        (Quantity_Interval_Disjoint(Quantity_Interval(0.0 * P.V, 1.0 * P.V)), None),
        (EnumSet(_Enum.A, _Enum.B), None),
        (EnumSet.unbounded(_Enum), None),
    ],
)
def test_serialize(input_set: P_Set, expected: P_Set | None):
    if expected is None:
        expected = input_set
    serialized = input_set.serialize()
    deserialized = P_Set.deserialize(serialized)
    assert deserialized == expected


@pytest.mark.parametrize(
    "value,expected",
    [
        (1.5, 2),
        (1, 1),
        (math.inf, math.inf),
        (-math.inf, -math.inf),
        (
            Quantity_Interval_Disjoint(Quantity_Interval(1.5, 2.5)),
            Quantity_Interval_Disjoint(Quantity_Interval(2, 2)),
        ),
        (
            Quantity_Interval_Disjoint(Quantity_Interval(1.5, math.inf)),
            Quantity_Interval_Disjoint(Quantity_Interval(2, math.inf)),
        ),
    ],
)
def test_float_round(value, expected):
    assert float_round(value) == expected


def test_regression_ss_zero():
    x = Quantity_Set_Discrete(2.77e-17)
    y = Quantity_Interval_Disjoint(Quantity_Interval(-math.inf, 0))
    assert x.is_subset_of(y)


@pytest.mark.parametrize(
    "digits,expected",
    [
        (0, Quantity_Interval_Disjoint(Quantity_Interval(2, 2))),
        (1, Quantity_Interval_Disjoint(Quantity_Interval(1.5, 2.4))),
        (2, Quantity_Interval_Disjoint(Quantity_Interval(1.51, 2.42))),
    ],
)
def test_round_digits(digits: int, expected: Quantity_Interval_Disjoint):
    x = Quantity_Interval_Disjoint(Quantity_Interval(1.51, 2.42))
    assert round(x, digits) == expected


@pytest.mark.parametrize(
    "value,digits,expected",
    [
        (1234.5678, 2, 1200),
        (1234.5678, 4, 1235),
        (1234.5678, 6, 1234.57),
        (1234.5678, 10, 1234.5678),
        (0.123456, 2, 0.12),
        (0.123456, 4, 0.1235),
        (0.123456, 6, 0.123456),
        (0.123456, 10, 0.123456),
    ],
)
def test_rel_round(value: float | int, digits: int, expected: float | int):
    assert rel_round(value, digits) == expected



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module


class TXS0102DCUR_UART(Module):
    """
    TXS0102DCUR 2 bit level shifter with UART interfaces
    - Output enabled by default
    """

    voltage_a_power: F.ElectricPower
    voltage_b_power: F.ElectricPower
    voltage_a_bus: F.UART_Base
    voltage_b_bus: F.UART_Base

    buffer: F.TXS0102DCUR

    def __preinit__(self):
        self.voltage_a_power.connect(self.buffer.voltage_a_power)
        self.voltage_b_power.connect(self.buffer.voltage_b_power)

        # enable output by default
        self.buffer.n_oe.set(True)

        # connect UART interfaces to level shifter
        self.voltage_a_bus.rx.connect(self.buffer.shifters[0].io_a)
        self.voltage_a_bus.tx.connect(self.buffer.shifters[1].io_a)
        self.voltage_b_bus.rx.connect(self.buffer.shifters[0].io_b)
        self.voltage_b_bus.tx.connect(self.buffer.shifters[1].io_b)

        # TODO
        # self.add(
        #    can_bridge_defined(self.voltage_a_bus, self.voltage_b_bus)
        # )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class has_simple_value_representation_defined(F.has_simple_value_representation.impl()):
    def __init__(self, value: str) -> None:
        super().__init__()
        self.value = value
        assert value != ""

    def get_value(self) -> str:
        return self.value



import pytest

from faebryk.libs.util import write_only_property


def test_write_only_property():
    """Test that write-only property raises on get and allows set"""

    class TestClass:
        def __init__(self):
            self._value = None

        @write_only_property
        def write_only(self, value):
            self._value = value

        def get_value(self):
            return self._value

    obj = TestClass()

    # Reading should raise AttributeError
    with pytest.raises(AttributeError) as exc_info:
        _ = obj.write_only
    assert "write_only is write-only" in str(exc_info.value)

    # Writing should work
    test_value = "test"
    obj.write_only = test_value
    assert obj.get_value() == test_value



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import inspect
import logging
from typing import Callable, Iterable, cast

from more_itertools import first

import faebryk.library._F as F
from atopile import errors
from faebryk.core.cpp import Path
from faebryk.core.graph import Graph, GraphFunctions
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.exceptions import accumulate
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class ERCFault(errors.UserException):
    """Base class for ERC faults."""


class ModuleInterfacePath(list[ModuleInterface]):
    """A path of ModuleInterfaces."""

    @classmethod
    def from_path(cls, path: Path) -> "ModuleInterfacePath":
        """
        Convert a Path (of GraphInterfaces) to a ModuleInterfacePath.
        """
        mifs = cast(
            list[ModuleInterface],
            [gif.node for gif in path if gif.node.isinstance(ModuleInterface)],
        )
        return cls(mifs)

    def snip_head(
        self, scissors: Callable[[ModuleInterface], bool], include_last: bool = True
    ) -> "ModuleInterfacePath":
        """
        Keep the head until the scissors predicate returns False.
        """
        for i, mif in enumerate(self):
            if not scissors(mif):
                return type(self)(self[: i + include_last])
        return self

    def snip_tail(
        self, scissors: Callable[[ModuleInterface], bool], include_first: bool = True
    ) -> "ModuleInterfacePath":
        """
        Keep the tail until the scissors predicate returns False.
        """
        for i, mif in reversed(list(enumerate(self))):
            if not scissors(mif):
                return type(self)(self[i + (not include_first) :])
        return self

    @classmethod
    def from_connection(
        cls, a: ModuleInterface, b: ModuleInterface
    ) -> "ModuleInterfacePath | None":
        """
        Return a ModuleInterfacePath between two ModuleInterfaces, if it exists,
        else None.
        """
        if paths := a.is_connected_to(b):
            # FIXME: Notes: from the master of graphs:
            #  - iterate through all paths
            #  - make a helper function
            #    ModuleInterfacePath.get_subpaths(path: Path, search: SubpathSearch)
            #    e.g SubpathSearch = tuple[Callable[[ModuleInterface], bool], ...]
            #  - choose out of subpaths
            #    - be careful with LinkDirectDerived edges (if there is a faulting edge
            #      is derived, save it as candidate and only yield it if no other found)
            #    - choose first shortest
            return cls.from_path(first(paths))
        return None


class ERCFaultShort(ERCFault):
    """Exception raised for short circuits."""


class ERCFaultShortedModuleInterfaces(ERCFaultShort):
    """Short circuit between two ModuleInterfaces."""

    def __init__(self, msg: str, path: ModuleInterfacePath, *args: object) -> None:
        super().__init__(msg, path, *args)
        self.path = path

    @classmethod
    def from_path(cls, path: ModuleInterfacePath) -> "ERCFaultShortedModuleInterfaces":
        """
        Given two shorted ModuleInterfaces, return an exception that describes the
        narrowest path for the fault.
        """
        return cls(f"`{" ~ ".join(mif.get_full_name() for mif in path)}`", path)


class ERCFaultElectricPowerUndefinedVoltage(ERCFault):
    def __init__(self, faulting_EP: F.ElectricPower, *args: object) -> None:
        msg = (
            f"ElectricPower with undefined or unsolved voltage: {faulting_EP}:"
            f" {faulting_EP.voltage}"
        )
        super().__init__(msg, [faulting_EP], *args)


class ERCPowerSourcesShortedError(ERCFault):
    """
    Multiple power sources shorted together
    """


def simple_erc(G: Graph, voltage_limit=1e5 * P.V):
    """Simple ERC check.

    This function will check for the following ERC violations:
    - shorted ElectricPower
    - shorted Caps
    - shorted Resistors
    - shorted symmetric footprints
    - shorted Nets
    - Net name collision

    - [unmapped pins for footprints]

    Args:

    Returns:
    """
    logger.info("Checking graph for ERC violations")

    with accumulate(ERCFault) as accumulator:
        # shorted power
        electricpower = GraphFunctions(G).nodes_of_type(F.ElectricPower)
        logger.info(f"Checking {len(electricpower)} Power")

        buses_grouped = ModuleInterface._group_into_buses(electricpower)
        buses = list(buses_grouped.values())

        # We do collection both inside and outside the loop because we don't
        # want to continue the loop if we've already raised a short exception
        with accumulator.collect():
            logger.info("Checking for hv/lv shorts")
            for ep in electricpower:
                if mif_path := ModuleInterfacePath.from_connection(ep.lv, ep.hv):

                    def _keep_head(x: ModuleInterface) -> bool:
                        if parent := x.get_parent():
                            parent_node, _ = parent
                            if isinstance(parent_node, F.ElectricPower):
                                return parent_node.hv is not x

                        return True

                    def _keep_tail(x: ModuleInterface) -> bool:
                        if parent := x.get_parent():
                            parent_node, _ = parent
                            if isinstance(parent_node, F.ElectricPower):
                                return parent_node.lv is not x

                        return True

                    raise ERCFaultShortedModuleInterfaces.from_path(
                        mif_path.snip_head(_keep_head).snip_tail(_keep_tail)
                    )

            logger.info("Checking for power source shorts")
            for bus in buses:
                with accumulator.collect():
                    sources = {
                        ep for ep in bus if ep.has_trait(F.Power.is_power_source)
                    }
                    if len(sources) <= 1:
                        continue

                    friendly_sources = ", ".join(n.get_full_name() for n in sources)
                    raise ERCPowerSourcesShortedError(
                        f"Power sources shorted: {friendly_sources}"
                    )

        # shorted nets
        nets = GraphFunctions(G).nodes_of_type(F.Net)
        logger.info(f"Checking {len(nets)} explicit nets")
        for net in nets:
            with accumulator.collect():
                collisions = {
                    p[0]
                    for mif in net.part_of.get_connected()
                    if (p := mif.get_parent()) and isinstance(p[0], F.Net)
                }

                if collisions:
                    friendly_shorted = ", ".join(n.get_full_name() for n in collisions)
                    raise ERCFaultShort(
                        f"Shorted nets: {friendly_shorted}",
                    )

        # shorted components
        # parts = [n for n in nodes if n.has_trait(has_footprint)]
        # sym_fps = [
        #    n.get_trait(has_footprint).get_footprint()
        #    for n in parts
        #    if n.has_trait(can_attach_to_footprint_symmetrically)
        # ]
        # logger.info(f"Checking {len(sym_fps)} symmetric footprints")
        # for fp in sym_fps:
        #    mifs = set(fp.get_all())
        #    checked = set()
        #    for mif in mifs:
        #        checked.add(mif)
        #        if any(mif.is_connected_to(other) for other in (mifs - checked)):
        #            raise ERCFault([mif], "shorted symmetric footprint")

        comps = GraphFunctions(G).nodes_of_types((F.Resistor, F.Capacitor, F.Fuse))
        logger.info(f"Checking {len(comps)} passives")
        for comp in comps:
            with accumulator.collect():
                assert isinstance(comp, (F.Resistor, F.Capacitor, F.Fuse))
                # TODO make prettier
                if (
                    comp.has_trait(F.has_part_picked)
                    and comp.get_trait(F.has_part_picked).removed
                ):
                    continue

                if path := ModuleInterfacePath.from_connection(
                    comp.unnamed[0], comp.unnamed[1]
                ):
                    raise ERCFaultShortedModuleInterfaces.from_path(path)

        ## unmapped Electricals
        # fps = [n for n in nodes if isinstance(n, Footprint)]
        # logger.info(f"Checking {len(fps)} footprints")
        # for fp in fps:
        #    for mif in fp.get_all():
        #        if not mif.get_direct_connections():
        #            raise ERCFault([mif], "no connections")

        # TODO check multiple pulls per logic


def check_modules_for_erc(module: Iterable[Module]):
    for m in module:
        logger.info(f"Checking {m} {'-'*20}")
        simple_erc(m.get_graph())


def check_classes_for_erc(classes: Iterable[Callable[[], Module]]):
    modules = []
    for c in classes:
        try:
            m = c()
        except Exception as e:
            logger.warning(
                f"Could not instantiate {c.__name__}: {type(e).__name__}({e})"
            )
            continue
        modules.append(m)
    check_modules_for_erc(modules)


def check_library_for_erc(lib):
    members = inspect.getmembers(lib, inspect.isclass)
    module_classes = [m[1] for m in members if issubclass(m[1], Module)]
    check_classes_for_erc(module_classes)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


# TODO this seems like it should be doing more
class USB2_0_ESD_Protection(Module):
    """
    USB 2.0 ESD protection
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    usb = L.list_field(2, F.USB2_0)

    vbus_esd_protection = L.p_field(domain=L.Domains.BOOL())
    data_esd_protection = L.p_field(domain=L.Domains.BOOL())

    no_pick: F.has_part_removed

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.usb[0], self.usb[1])

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------
        self.usb[0].connect(self.usb[1])
        self.usb[0].usb_if.buspower.connect(self.usb[1].usb_if.buspower)
        # FIXME
        # self.usb[0].usb_if.buspower.decoupled.decouple()

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        self.usb[0].usb_if.buspower.voltage.constrain_subset(
            L.Range(4.75 * P.V, 5.25 * P.V)
        )
